from __future__ import absolute_import, unicode_literals
from __future__ import print_function
from __future__ import division
from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      map, next, oct, pow, range, round,
                      str, super, zip)

import collections as c
import contextlib
import os
import os.path as op
import shutil
import subprocess as sp
import sys
import time
import math
import warnings
import itertools as it
import json
try:
    from json import JSONDecodeError #py3k
except ImportError: #py2k
    JSONDecodeError = ValueError
try:
    import io
except ImportError:  # Python 2
    import StringIO as io
try:
    import shutil
    which = shutil.which
except (ImportError, AttributeError):
    import distutils.spawn
    which = distutils.spawn.find_executable

from io import open
import logging
from pprint import pprint


import numpy as np

import scipy.spatial
import scipy.stats

import Bio.PDB as bpdb

from logging_exceptions import log_to_exception, log_exception

from ...graph import bulge_graph as fgb
import forgi.graph.residue as fgr
from ...graph.sequence import Sequence, _insert_breakpoints_simple
from ...graph._graph_construction import _BulgeGraphConstruction
from ..utilities import graph_pdb as ftug
from . import stats as ftms
from . import transform_cg as ftmt
from ..._k2n_standalone import knotted2nested as cak
from ..utilities import mcannotate as ftum
from ..utilities import pdb as ftup
from . import descriptors as ftmd
from ..utilities import _dssr as ftud
from ..utilities import vector as ftuv
from ...utilities import debug as fud
from ...utilities import stuff as fus
import forgi.threedee.utilities.virtual_residues as ftuvres
from ...utilities.observedDict import observedDict
from ...utilities.exceptions import CgConstructionError, CgIntegrityError, GraphConstructionError
from .linecloud import CoordinateStorage, LineSegmentStorage
from ...utilities.stuff import is_string_type
from ... import config

log = logging.getLogger(__name__)

class AnnotationToolNotInstalled(ValueError):
    pass


try:
    profile  # The @profile decorator from line_profiler (kernprof)
except:
    def profile(x):
        return x


def add_longrange_interactions(cg, lines):
    '''
    Iterate over the lines in an MC-Annotate file and add information
    about interactions between non-adjacent elements.

    :param cg: A CoarseGrainRNA structure
    :param lines: All the lines in an MC-Annotate file
    '''
    for line in ftum.iterate_over_interactions(lines):
        (from_chain, from_base, to_chain,
         to_base) = ftum.get_interacting_base_pairs(line)
        try:
            seq_id1 = cg.seq_ids.index(
                fgr.RESID(from_chain, ftum.parse_resid(from_base))) + 1
        except ValueError as e:
            with log_to_exception(log, e):
                log.error("seq_ids are {}".format(cg.seq_ids))
            raise
        seq_id2 = cg.seq_ids.index(
            fgr.RESID(to_chain, ftum.parse_resid(to_base))) + 1

        node1 = cg.get_node_from_residue_num(seq_id1)
        node2 = cg.get_node_from_residue_num(seq_id2)

        if abs(seq_id2 - seq_id1) > 1 and node1 != node2 and not cg.has_connection(node1, node2):
            cg.longrange[node1].add(node2)
            cg.longrange[node2].add(node1)


def breakpoints_from_seqids(seqids):
    """
    Create the list of cofold cutpoints from the seq_ids.
    Return 1-based breakpoints
    """
    breakpoints = []
    old_chain = None
    for i, seqid in enumerate(seqids):
        if old_chain is not None and seqid.chain != old_chain:
            assert i >= 1
            breakpoints.append(i)  # Breakpoint BEFORE current seqid!
        old_chain = seqid.chain
    return breakpoints


def _are_adjacent_basepairs(seq_ids, edge1, edge2):
    """
    Helper function used by connected_cgs_from_pdb.

    :param cg: A (potentially) inconsistent cg. Only seq_ids are needed.
    :param edge1, edge2: A dictionary with the key "basepair" mapping to a two-tuple of RESIDS
    """
    fromA, toA = edge1["basepair"]
    fromB, toB = edge2["basepair"]
    if fromA.chain != fromB.chain:
        if fromA.chain == toB.chain:
            fromB, toB = toB, fromB
            assert toA.chain == toB.chain
        else:
            assert False
    if (abs(seq_ids.index(fromA) - seq_ids.index(fromB)) == 1 and
            abs(seq_ids.index(toA) - seq_ids.index(toB)) == 1):
        log.debug(
            "Basepairs %s and %s are adjacent. No length 1 stem", edge1, edge2)
        return True
    log.debug("Basepairs %s and %s are NOT adjacent.", edge1, edge2)
    return False


def _annotate_pdb(filename, annotation_tool, filetype, subprocess_kwargs={}):
    """
    Get the secondary structure of the pdb by using an external tool.

    :param filename: The name of the temporary pdb file
    :param subprocess_kwargs: Will be passed as keyword arguments to subprocess.call/ subprcess.check_output
    :param annotation_tool: See docstring of CoarseGrainRNA.from_pdb
    """
    not_installed_msg = ("{} was requested as annotation tool, but {} "
                         "is not installed or not in the PATH. (Hint: run "
                         "forgi_config.py to set the preferred annotation tool)")
    program = annotation_tool
    if program is None:
        c = config.read_config()
        if "PDB_ANNOTATION_TOOL" in c:
            program = c["PDB_ANNOTATION_TOOL"]
    log.info("Requested annotation program is %s", program)
    if program == "DSSR" or program is None:
        if which("x3dna-dssr"):
            return _run_dssr(filename, subprocess_kwargs)
        else:
            log.info("x3dna-dssr is not installed or not in the PATH.")
            if program is not None:
                raise AnnotationToolNotInstalled(
                    not_installed_msg.format(program, "x3dna-dssr"))
    if program == "MC-Annotate" or (program is None and filetype == "pdb"):
        if filetype != "pdb":
            raise ValueError("MC-Annotate does not support MMCIF")
        if which("MC-Annotate"):
            lines = _run_mc_annotate(filename, subprocess_kwargs)
            try:
                bpseq, seq_ids = ftum.get_dotplot(lines)
                return bpseq, seq_ids, {}
            except Exception as e:
                log.exception(
                    "Could not convert MC-Annotate output to dotplot")
                raise CgConstructionError(
                    "Could not convert MC-Annotate output to dotplot")  # from e
        else:
            log.info("MC-Annotate is not installed or not in the PATH.")
            if program is not None:
                raise AnnotationToolNotInstalled(not_installed_msg.format(
                    program, "MC-Annotate"))
    elif program == "forgi" or program is None:
        return None
    else:
        raise ValueError("Supported programs for annotating the pdb are: "
                         "'MC-Annotate', 'DSSR' and 'forgi', not '{}'".format(program))


def _run_dssr(filename, subprocess_kwargs={}):
    # See http://forum.x3dna.org/rna-structures/redirect-auxiliary-file-output/
    with fus.make_temp_directory() as dssr_output_dir:
        dssr_out = os.path.join(dssr_output_dir, "out")

        arguments = ['x3dna-dssr', "-i=" + filename,
                     "--prefix=" + os.path.join(dssr_output_dir, "d"),
                     "-o=" + dssr_out, "--json"]
        # https://stackoverflow.com/a/14837250/5069869
        log.info("Running DSSR: %s", sp.list2cmdline(arguments))
        with open(os.path.join(dssr_output_dir, "stderror"), "w+") as errfile:
            try:
                ret_code = sp.call(
                    arguments, universal_newlines=True, stderr=errfile, **subprocess_kwargs)
                try:
                    with open(os.path.join(dssr_output_dir, "d-2ndstrs.bpseq"), encoding='ascii') as f:
                        bpseq = f.read()
                    with open(dssr_out) as f:
                        try:
                            dssr_dict = json.load(f)
                        except JSONDecodeError as e: # DSSR does not escape backslashes (see "http://forum.x3dna.org/bug-reports/json-output-should-escape-backslashes/"), which can be in windows paths. Escape them manually.
                            f.seek(0)
                            json_stri = f.read().replace('\\','\\\\')
                            log.error(json_stri)
                            dssr_dict = json.loads(json_stri)
                        nts = dssr_dict["nts"]
                        seq_ids = list(map(ftud.dssr_to_pdb_resid, [
                                       nt["nt_id"] for nt in nts]))
                except (OSError, IOError) as e:
                    with log_to_exception(log, e):
                        log.error("Content of the directory %s is %s",
                                  dssr_output_dir, os.listdir(dssr_output_dir))
                    raise
            except (OSError, IOError) as e:
                assert op.isfile(
                    filename), "File {} (created by forgi) no longer exists".format(filename)
                errfile.seek(0)
                err_msg = errfile.readlines()
                if err_msg:
                    if len(err_msg) >= 3:
                        e = CgConstructionError(
                            "DSSR could not process the file: " + err_msg[-3])
                    with log_to_exception(log, e):
                        log.error("Captured Stderr is:\n%s", "".join(
                            ["... " + line for line in err_msg]))
                    raise e
                else:
                    # On Py2.7, the filename of the executable (x3dna-dssr) is not part of the
                    # error message raised by sp.call,
                    # For this reason, we add the following hint (unneeded on Py3k)
                    e.strerror += ". (Hint: Did you install x3dna-dssr?)"
                raise e
    return bpseq, seq_ids, dssr_dict


def _run_mc_annotate(filename, subprocess_kwargs={}):
    """
    Returns: A tuple (bpseq, seq_ids`)
    """
    log.info("Running MC-Annotate")
    try:
        out = sp.check_output(['MC-Annotate', filename],
                              universal_newlines=True, **subprocess_kwargs)
    except OSError as e:
        assert op.isfile(
            filename), "File {} (created by forgi) no longer exists".format(filename)
        e.strerror += ". Hint: Did you install MC-Annotate?"
        raise e

    lines = out.strip().split('\n')
    return lines


class RnaMissing3dError(LookupError):
    pass


class CoarseGrainRNA(fgb.BulgeGraph):
    '''
    A coarse grain model of RNA structure based on the
    bulge graph representation.

    Each stem is represented by four parameters (two endpoints)
    and two twist vetors pointing towards the centers of the base
    pairs at each end of the helix.
    '''

    def __init__(self, graph_construction, sequence, name=None, infos=None, _dont_split=False):
        '''
        Initialize the new structure.
        '''
        super(CoarseGrainRNA, self).__init__(
            graph_construction, sequence, name, infos, _dont_split)

        self.dssr = None

        self._virtual_atom_cache = {}
        #: Keys are element identifiers (e.g.: "s1" or "i3"), values are 2-tuples of vectors
        #: The first value of stem coordinates corresponds to the start of the stem
        #: (the one with the lowest nucleotide number),
        #: The second value to the end of the stem.
        #: If the coordinates for an element change, the virtual atom and virtual residue
        #: coordinates are automatically invalidated.
        self.coords = None  # We can only initialize this, when we know defines.keys()
        self.twists = None
        self.sampled = dict()

        if self.defines:
            self._init_coords()

        #: A dictionary storing the stem-bases (not the vres basis)
        self.bases = {}
        self.stem_invs = {}


        #: generated by self.add_all_virtual_residues()
        self.vposs = c.defaultdict(dict)
        self.vbase = c.defaultdict(dict)
        self.vsugar = c.defaultdict(dict)
        self.vbackbone = c.defaultdict(dict)
        #:The following defaultdicts are cleared when coords or twists change.
        #: The coordinate system specific to each virtual residue
        #: (3x3 matrix, carthesian coordiantes)
        #: generated by self.add_all_virtual_residues()
        self.vbases = c.defaultdict(dict)
        #: generated by self.add_all_virtual_residues()
        self.vvecs = c.defaultdict(dict)
        #: generated by self.add_all_virtual_residues()
        self.v3dposs = c.defaultdict(dict)
        #: generated by self.add_all_virtual_residues()
        #: The inverse of the transposed basis.
        self.vinvs = c.defaultdict(dict)

        #: A 3D vector. Used as hint, from what direction the Projection2D object
        #: should be generated in the default case.
        self.project_from = None

        self.longrange = c.defaultdict(set)
        self.chains = {}  # the PDB chains if loaded from a PDB file

        self.interacting_residues = []

        # Lazily loaded:
        self._incomplete_elements = None
        # Lazily calculated from interacting_residues
        self._interacting_elements = None
    ############################################################################
    # Factory functions
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    @classmethod
    def from_bg_string(cls, cg_string):
        '''
        Populate this structure from the string
        representation of a graph.
        '''
        # pylint: disable=no-member  # See https://github.com/PyCQA/pylint/issues/981
        # Reading the bulge_graph-part of the file
        cg = super(CoarseGrainRNA, cls).from_bg_string(cg_string)
        cg._init_coords()
        log.debug("BG read. Now reading 3D information")
        # Reading the part of the file responsible for 3D information
        lines = cg_string.split('\n')
        for line in lines:
            line = line.strip()
            parts = line.split()
            if len(parts) == 0:
                continue
            if parts[0] == 'coord':
                name = parts[1]
                cg.coords[name] = np.array(
                    [list(map(float, parts[2:5])), list(map(float, parts[5:8]))])
            if parts[0] == 'twist':
                name = parts[1]
                cg.twists[name] = np.array(
                    [list(map(float, parts[2:5])), list(map(float, parts[5:8]))])
            if parts[0] == 'longrange':
                cg.longrange[parts[1]].add(parts[2])
                cg.longrange[parts[2]].add(parts[1])

            if parts[0] == 'sampled':
                cg.sampled[parts[1]] = [parts[2]] + list(map(int, parts[3:]))
            if parts[0] == 'project':
                cg.project_from = np.array(parts[1:], dtype=float)
            if parts[0] == "interacting":
                cg.interacting_residues.append(fgr.resid_from_str(parts[1]))
            if parts[0] == "vres":
                elem = parts[1]
                cg.vposs[elem] = ftuvres.parse_vres(parts[2:])[0]
            if parts[0] == "vbase":
                elem = parts[1]
                cg.vbase[elem] = ftuvres.parse_vres(parts[2:])[0]
            if parts[0] == "vsugar":
                elem = parts[1]
                cg.vsugar[elem] = ftuvres.parse_vres(parts[2:])[0]
            if parts[0] == "vbackbone":
                elem = parts[1]
                cg.vbackbone[elem] = ftuvres.parse_vres(parts[2:])[0]
        # Old versions of the file may contain bulge coordinates in the wrong order.
        cg.add_bulge_coords_from_stems()
        return cg

    @classmethod
    def from_pdb(cls, pdb_filename, load_chains=None, remove_pseudoknots=False,
                 dissolve_length_one_stems=True, secondary_structure=None,
                 filetype="pdb", annotation_tool=None, query_PDBeChem=False):
        """
        :param load_chains: A list of chain_ids or None (all chains)
        :param secondary_structure: Only useful if we load only 1 component
        :param filetype: One of 'pdb' or 'cif'
        :param query_PDBeChem: If true, query the PDBeChem database whenever a
                        modified residue with unknown 3-letter code
                        is encountered.
        :param annotation_tool: One of "DSSR", "MC-Annotate", "forgi" or None.
                        If this is None, we take the value of the configuration
                        file (run forgi_config.py to create a config file).
                        If no config file is given either, we see what tools are
                        installed, preferring the newer DSSR over MC-Annotate and
                        falling back to the fogi implementation if neither is in
                        the PATH variable.
                        If a string is given or the configuration file set,
                        we never fall back to a different option but raise an
                        error, if the requested tool is unavailable.
        """
        #warnings.warn("We currently do not load any long-range interactions")
        # We need to create files, so we can interface with
        # the annotation program.
        if remove_pseudoknots and secondary_structure:
            warnings.warn(
                "Option 'remove_pseudoknots ignored for user-supplied secondary structure.")

        with fus.make_temp_directory() as output_dir:
            if load_chains == "biggest":
                chain, missing_res, ir = ftup.get_biggest_chain(pdb_filename)
                chains = [chain]
            else:
                chains, missing_res, ir = ftup.get_all_chains(pdb_filename)
            new_chains = []
            for chain in chains:
                if load_chains in [None, "biggest"] or chain.id in load_chains:
                    log.debug("Loaded Chain %s", chain.id)
                    chain, modifications = ftup.clean_chain(chain, query_PDBeChem)
                    new_chains.append(chain)
            log.debug("%s, %s", pdb_filename, os.path.split(pdb_filename))
            fn_basename = os.path.split(pdb_filename)[1]
            if load_chains is None:
                chain_string = "None"
            else:
                chain_string = "-".join(map(str, load_chains))

            rna_pdb_fn = op.join(output_dir, fn_basename +
                                 "_" + chain_string + '.temp.' + filetype)
            # Output cleaned version for annotating
            ftup.output_multiple_chains(new_chains, rna_pdb_fn, filetype)

            # first we annotate the 3D structure
            log.info("Starting annotation program for all chains")
            annotation = _annotate_pdb(rna_pdb_fn, annotation_tool, filetype)
            if annotation is None:
                # Fallback-annotation using forgi
                bpseq, seq_ids = ftup.annotate_fallback(new_chains)
                dssr_dict = {}
            else:
                bpseq, seq_ids, dssr_dict = annotation
            breakpoints = breakpoints_from_seqids(seq_ids)

        assert bpseq is not None, "{} {} {}".format(bpseq, seq_ids, dssr_dict)
        # Here, we remove Pseudoknots on the bp-seq level. not the BulgeGraph
        # level, because there may be multiple components, so we cannot create
        # a BulgeGraph.
        if remove_pseudoknots:
            #log.info("Removing pseudoknots")
            #log.info(bpseq)
            bpseq = cak.k2n_main(io.StringIO(bpseq), input_format='bpseq',
                                 #output_format = 'vienna',
                                 output_format='bpseq',
                                 method=cak.DEFAULT_METHOD,
                                 opt_method=cak.DEFAULT_OPT_METHOD,
                                 verbose=cak.DEFAULT_VERBOSE,
                                 removed=cak.DEFAULT_REMOVED)

        import networkx as nx
        chain_connections_multigraph = nx.MultiGraph()
        bpseq_lines = bpseq.splitlines()
        for line in bpseq_lines:
            try:
                from_, res, to_ = line.split()
            except:
                continue
            from_seqid = seq_ids[int(from_) - 1]
            from_chain = from_seqid.chain
            chain_connections_multigraph.add_node(from_chain)
            if int(to_) != 0 and int(to_) > int(from_):
                to_seqid = seq_ids[int(to_) - 1]
                to_chain = to_seqid.chain
                if from_chain != to_chain:
                    log.debug("Adding {} - {}".format(from_chain, to_chain))
                    chain_connections_multigraph.add_edge(
                        from_chain, to_chain, basepair=(from_seqid, to_seqid))
        chain_connections = nx.Graph(chain_connections_multigraph)
        if dissolve_length_one_stems:
            for chain1, chain2 in it.combinations(chain_connections_multigraph.nodes(), 2):
                if chain2 in chain_connections_multigraph.adj[chain1]:
                    for edge1, edge2 in it.combinations(chain_connections_multigraph.adj[chain1][chain2].values(), 2):
                        if _are_adjacent_basepairs(seq_ids, edge1, edge2):
                            break
                    else:  # break NOT encountered.
                        log.debug(
                            "No connection remaining between chains %s and %s", chain1, chain2)
                        chain_connections.remove_edge(chain1, chain2)

        log.debug("CONNECTIONS: {}, nodes {}".format(
            chain_connections, chain_connections.nodes()))
        log.debug("Edges {}".format(chain_connections.edges()))

        pdb_base = op.splitext(op.basename(pdb_filename))[0]

        components = list(nx.connected_components(chain_connections))
        cgs = []
        if len(components) != 1 and secondary_structure:
            raise GraphConstructionError("User-provided secondary structure is "
                                         "ambigouse if more than 1 connected "
                                         "component loaded.")
        for component in nx.connected_components(chain_connections):
            try:
                cgs.append(cls._load_pdb_component(bpseq_lines, pdb_base, new_chains,
                                                   component, missing_res, modifications,
                                                   seq_ids, secondary_structure,
                                                   dissolve_length_one_stems,
                                                   ir))
            except GraphConstructionError as e:
                log_exception(e, logging.ERROR, with_stacktrace=False)
                log.error(
                    "Could not load chains %s, due to the above mentioned error.", list(component))
                raise

        cgs.sort(key=lambda x: x.name)
        if dssr_dict:
            for cg in cgs:
                cg.dssr = ftud.DSSRAnnotation(dssr_dict, cg)
                cg.infos["dssr_stacks"] = cg.dssr.stacking_loops()
        log.debug("Returning %s", cgs)
        return cgs

    @classmethod
    def _load_pdb_component(cls, original_bpseq_lines, name, chains, chain_ids,
                            missing_res, modifications, seq_ids,
                            secondary_structure="", dissolve_length_one_stems=False,
                            interacting_residues=[]):
        """
        :param original_bpseq_lines: List of strings. Will be filtered for chains.
        """
        #print(component, type(component))
        log.info("Loading PDB: Connected component with chains %s",
                 str(list(chain_ids)))
        log.debug("missing residues %s", missing_res)
        # Since the external annotation program can take some time,
        # we do not re-annotate, but filter the bpseq_string instead.
        new_bpseq = []
        new_seqids = []

        # Filter the bpseq_lines
        for line in original_bpseq_lines:
            from_, res, to_ = line.split()
            from_seqid = seq_ids[int(from_) - 1]
            if from_seqid.chain in chain_ids:
                new_seqids.append(from_seqid)
                if to_ != '0':
                    to_seqid = seq_ids[int(to_) - 1]
                    if to_seqid.chain not in chain_ids:  # Because of length-1-basepair
                        to_ = 0
                new_bpseq.append((from_, res, to_))

        new_bpseq_str = fus.renumber_bpseq(new_bpseq)
        tuples, seq_str = fus.bpseq_to_tuples_and_seq(new_bpseq_str)
        breakpoints = breakpoints_from_seqids(new_seqids)
        seq_str = _insert_breakpoints_simple(seq_str, breakpoints, 1)

        if secondary_structure:
            pt = fus.dotbracket_to_pairtable(secondary_structure)
            # Overwrite tuples!
            tuples = fus.pairtable_to_tuples(pt)
            stru_fragments = secondary_structure.split("&")
            seq_fragments = seq_str.split("&")
            seq_lengths = list(map(len, seq_fragments))
            if list(map(len, stru_fragments)) != seq_lengths:
                raise ValueError("The given secondary structure is inconsistent "
                                 "with the pdb-chain-lengths of %s", seq_lengths)
        sequence = Sequence(seq_str, new_seqids,
                            [r for r in missing_res if r["chain"] in chain_ids],
                            {k: v for k, v in modifications.items() if k.chain in chain_ids})

        name = name + "_" + "-".join(c for c in sorted(chain_ids))
        graph_constr = _BulgeGraphConstruction(tuples)
        cg = cls(graph_constr, sequence, name)
        cg = fgb._cleaned_bg(cg, dissolve_length_one_stems)
        cg.chains = {
            chain.id: chain for chain in chains if chain.id in chain_ids}

        ftug.add_stem_information_from_pdb_chains(cg)
        cg.add_bulge_coords_from_stems()
        ftug.add_loop_information_from_pdb_chains(cg)
        ftug._add_loop_vres(cg)
        assert len(cg.defines) == len(
            cg.coords), cg.defines.keys() ^ cg.coords.keys()
        cg.interacting_residues = list(r for r in map(fgr.resid_from_biopython, interacting_residues)
                                       if r.chain in chain_ids)
        return cg

    ############################################################################

    @property
    def transformed(self):
        return ftmt.CGTransformer(self)

    def _get_coord_str(self):
        '''
        Place the start and end coordinates of each stem into a string.

        The format is:

            coord s1 x1 y1 z1 x2 y2 z2

        Where s1 is the name of the stem, (x1,y1,z1) and (x2,y2,z2) are
        the two endpoints of the stem.

        :return: A string containing the coordinates for all of the stems.
        '''

        out_str = ''
        for key in self.coords.keys():
            if key[0] in ["m", "i"]:
                # Bulge coordinates are redundant. They can be deduced from the stem coordinates.
                continue
            [p, n] = self.coords[key]
            out_str += ("coord {k} {x[0]:.16f} {x[1]:.16f} {x[2]:.16f} "
                        "{y[0]:.16f} {y[1]:.16f} {y[2]:.16f}".format(k=key, x=p, y=n))
            out_str += '\n'
        return out_str

    def add_bulge_coords_from_stems(self):
        '''
        Add the information about the starts and ends of the bulges (i and m elements).
        The stems have to be created beforehand.

        This is called during loading of the RNA structure from pdb and from cg files.
        '''
        for d in self.defines.keys():
            if d[0] != 's':
                edges = list(self.edges[d])
                if len(edges) == 2:
                    (s1b, _) = self.get_sides(edges[0], d)
                    (s2b, _) = self.get_sides(edges[1], d)

                    mids1 = self.coords[edges[0]]
                    mids2 = self.coords[edges[1]]

                    # Save coordinates in direction of the strand.
                    if self.get_link_direction(edges[0], edges[1], d) == 1:
                        self.coords[d] = (mids1[s1b], mids2[s2b])
                    else:
                        self.coords[d] = (mids2[s2b], mids1[s1b])

    def add_all_virtual_residues(self):
        """
        Calls ftug.add_virtual_residues() for all stems of this RNA.

        .. note::
           Don't forget to call this again if you changed the structure of the RNA,
           to avoid leaving it in an inconsistent state.

        .. warning::
           Virtual residues are only added to stems, not to loop regions.
           The position of residues in loops is much more flexible, which is why virtual
           residue positions for loops usually do not make sense. If the file
           was loaded from the PDB, residue positions from the PDB file are
           stored already.
        """
        for stem in self.stem_iterator():
            try:
                log.debug(
                    "Adding virtual residues for stem %s with coords %s", stem, self.coords[stem])
                ftug.add_virtual_residues(self, stem)
            except (KeyError, ValueError, AssertionError):
                if np.all(np.isnan(self.coords[stem])):
                    raise RnaMissing3dError(
                        "No 3D coordinates available for stem {}".format(stem))
                elif stem[0] == "s" and np.all(np.isnan(self.twists[stem])):
                    raise RnaMissing3dError(
                        "No twists available for stem {}".format(stem))
                else:
                    log.info("Reraising an ERROR in add_all_virtual_residues")
                    raise

    def iter_three_points(self, pos):
      elem = self.get_node_from_residue_num(pos)
      d = self.defines[elem]
      if d[0] <= pos <= d[1]:
          i = pos - d[0]
      else:
          i = d[1] - d[0] + 1 + pos - d[2]
      for coord_type in self.vbase, self.vsugar, self.vbackbone:
          elem_coords = coord_type[elem][i]
          origin, basis = ftug.element_coord_system(self, elem)
          new_coords = ftuv.change_basis(
              elem_coords, ftuv.standard_basis, basis) + origin
          log.debug("%s %s: coords %s mapped to %s", elem, i, elem_coords, new_coords)
          yield new_coords

    def get_virtual_residue(self, pos, allow_single_stranded=False):
        """
        Get the virtual residue position in the global coordinate system
        for the nucleotide at position pos (1-based)

        :param pos: A 1-based nucleotide number
        :param allow_single_stranded: If True and pos is not in a stem, return a
              rough estimate for the residue position instead of raising an error.
              Currenly, for non-stem elements, these positions are on the axis of the cg-element.
        """
        mult = 5
        if isinstance(pos, fgr.RESID):
            pos = self.seq.to_integer(pos)
        elem = self.get_node_from_residue_num(pos)
        if elem[0] == "s" and elem not in self.v3dposs or not self.v3dposs[elem]:
            ftug.add_virtual_residues(self, elem)
        if elem[0] == "s":
            for i in range(self.stem_length(elem)):
                if self.defines[elem][0] + i == pos:
                    vres = self.v3dposs[elem][i]
                    return vres[0] + vres[3] * mult
                elif self.defines[elem][3] - i == pos:
                    vres = self.v3dposs[elem][i]
                    return vres[0] + vres[2] * mult
            else:
                assert False
        else:
            d = self.defines[elem]
            if d[0] <= pos <= d[1]:
                i = pos - d[0]
            else:
                i = d[1] - d[0] + 1 + pos - d[2]
            try:
                elem_coords = self.vposs[elem][i]
                origin, basis = ftug.element_coord_system(self, elem)
                new_coords = ftuv.change_basis(
                    elem_coords, ftuv.standard_basis, basis) + origin
                log.debug("%s %s: coords %s mapped to %s", elem, i, elem_coords, new_coords)
                return new_coords
            except KeyError as e:
                try:
                    self._has_warned_old_vres
                except AttributeError:
                    self._has_warned_old_vres = set()
                if elem not in self._has_warned_old_vres:
                    log.warning(
                        "No virtual residues have been loaded for loops: %s (for pos %s) in elem %s."
                        "Using inaccurate position along the cylinder instead.", e, pos, elem)
                    self._has_warned_old_vres.add(elem)
                if not allow_single_stranded:
                    raise ValueError(
                        "Position {} is not in a stem! It is in {}.".format(pos, elem))
                if elem[0] == "h":
                    # We estimate the vres position along the axis of the hairpin.
                    h_length = self.element_length(elem) / 2
                    pos_in_h = pos - self.defines[elem][0] +1
                    if pos_in_h > math.ceil(h_length):
                        l = self.defines[elem][1] - pos+1
                    elif pos_in_h <=int(h_length):
                        l = pos - self.defines[elem][0]+1
                    else:
                        l = math.ceil(h_length)
                    perc = (l) / math.ceil(h_length)
                elif elem[0] == "i":
                    if pos <= self.defines[elem][1]:
                        l = pos - self.defines[elem][0]
                        tl = (self.defines[elem][1] - self.defines[elem][0])
                    else:
                        l = self.defines[elem][3] - pos
                        tl = (self.defines[elem][3] -
                              self.defines[elem][2] )
                    perc = (l+1) / (tl+2)
                else:
                    l = pos - self.defines[elem][0]
                    perc = (l+1) / (self.element_length(elem)+1)
                log.debug("Calculating vres: %s,%s, %s: length=%s, perc=%s",elem, pos, l,(self.coords[elem][1] - self.coords[elem][0]), perc)
                return self.coords[elem][0] + (self.coords[elem][1] - self.coords[elem][0]) * perc

    def get_ordered_stem_poss(self):
        points = []
        for s in self.sorted_stem_iterator():
            points += list(self.coords[s])
        points = np.array(points)
        if np.any(np.isnan(points)):
            raise RnaMissing3dError("No 3D coordinates present.")
        return points

    def get_ordered_virtual_residue_poss(self, return_elements=False):
        """
        Get the coordinates of all stem's virtual residues in a consistent order.

        This is used for RMSD calculation.
        If no virtual_residue_positions are known, self.add_all_virtual_residues() is called
        automatically.

        :param return_elements: In addition to the positions, return a list with
                                the cg-elements these coordinates belong to
        :returns: A numpy array.
        """
        if not self.v3dposs:
            self.add_all_virtual_residues()
        vress = []
        elems = []
        for i in range(1, len(self.seq)+1):
            pos = self.get_virtual_residue(i, allow_single_stranded = True)
            vress.append(pos)
            elems.append(self.get_elem(i))
        if return_elements:
            return np.array(vress), elems
        return np.array(vress)

    def get_poss_for_domain(self, elements, mode="vres"):
        """
        Get an array of coordinates only for the elements specified.

        ..note::

            This code is still experimental in the current version of forgi.

        :param elements: A list of coarse grain element names.
        """
        points = []
        if mode == "fast":
            for e in sorted(elements):
                points += list(self.coords[e])
            return np.array(points)
        elif mode == "vres":
            if not self.v3dposs:
                self.add_all_virtual_residues()
            vress = []
            for s in sorted(elements):
                if s[0] != "s":
                    continue
                for i in range(self.stem_length(s)):
                    vres = self.v3dposs[s][i]
                    vress += [vres[0] + vres[2], vres[0] + vres[3]]
            return np.array(vress)
        assert False

    def steric_value(self, elements, method="r**-2"):
        """
        Estimate, how difficult a set of elements was to build,
        by counting the atom density around the center of these elements
        """
        try:
            if isinstance(elements, list) or isinstance(elements, tuple):
                center = ftuv.get_vector_centroid(self.coords[elements])
            elif elements.shape == (3,):
                center = elements
            else:
                assert False, repr(elements)
        except:
            print(elements, repr(elements))
            raise
        if method == "kde":
            # print(center)
            all_vas = []
            for pos in range(1, self.seq_length + 1):
                for va in self.virtual_atoms(pos).values():
                    all_vas.append(va)

            all_vas = np.array(all_vas).T
            log.debug("Shape of all atoms {}".format(all_vas.shape))
            # randomly take 50 Angstrom bandwidth
            kde = scipy.stats.gaussian_kde(all_vas, 50)
            return kde(center)
        elif method.startswith("r**"):
            power = -int(method[3:5])
            exclude = method[5:]
            if exclude and exclude != "e":
                raise ValueError("Not supported method")
            value = 0
            for pos in range(1, self.seq_length + 1):
                if exclude and self.get_node_from_residue_num(pos) in elements:
                    continue
                for va in self.virtual_atoms(pos).values():
                    value += 1 / (1 + ftuv.vec_distance(va, center))**power
            return value
        elif method.startswith("cutoff"):
            cutoff = float(method.split()[1])
            value = 0
            for pos in range(1, self.seq_length + 1):
                for va in self.virtual_atoms(pos).values():
                    if ftuv.vec_distance(va, center) < cutoff:
                        value += 1
            return value

    def _get_twist_str(self):
        '''
        Place the twist vectors into a string.

        The format is:

            twist s1 x1 y1 z1 x2 y2 z2

        Where s1 is the name of the stem and (x1,y1,z1) and (x2,y2,z2) are
        the unit vectors from the start of each end of the stem to the middle
        of the base pairs.
        '''
        out_str = ''
        for key in self.twists.keys():
            [p, n] = self.twists[key]
            out_str += ("twist {k} {x[0]:.16f} {x[1]:.16f} {x[2]:.16f} "
                        "{y[0]:.16f} {y[1]:.16f} {y[2]:.16f}".format(k=key, x=p, y=n))
            out_str += '\n'
        return out_str

    def _get_long_range_str(self):

        out_str = ''
        printed = set()

        for key1 in self.longrange.keys():
            for key2 in self.longrange[key1]:
                k = [key1, key2]
                k.sort()
                k = tuple(k)
                if k in printed:
                    continue
                printed.add(k)

                out_str += "longrange %s %s\n" % (key1, key2)

        return out_str

    def _get_interacting_str(self):
        out = []
        for seqid in self.interacting_residues:
            out.append("interacting\t{}".format(fgr.resid_to_str(seqid)))
        return "\n".join(out) + "\n"

    def _get_sampled_stems_str(self):
        out_str = ''
        for key in self.sampled.keys():
            out_str += 'sampled %s %s\n' % (key,
                                            " ".join(map(str, self.sampled[key])))
        return out_str

    def _get_vres_str(self):
        out_str = ""
        for elem in self.vposs:
            if elem[0] != "s":
                out_str += "vres {} ".format(elem)
                out_str += ftuvres.serialize_vres(self.vposs[elem])
                out_str += "\n"
        for elem in self.vbase: 
            out_str += "vbase {} ".format(elem)
            out_str += ftuvres.serialize_vres(self.vbase[elem])
            out_str += "\n"
            out_str += "vsugar {} ".format(elem)
            out_str += ftuvres.serialize_vres(self.vsugar[elem])
            out_str += "\n"
            out_str += "vbackbone {} ".format(elem)
            out_str += ftuvres.serialize_vres(self.vbackbone[elem])
            out_str += "\n"
        return out_str

    def to_cg_string(self):
        '''
        Output this structure in string form.
        '''
        curr_str = self.to_bg_string()
        curr_str += self._get_coord_str()
        curr_str += self._get_twist_str()
        curr_str += self._get_sampled_stems_str()
        curr_str += self._get_long_range_str()
        curr_str += self._get_interacting_str()
        curr_str += self._get_vres_str()
        if self.project_from is not None:
            curr_str += "project {} {} {}\n".format(*self.project_from)
        return curr_str

    def to_file(self, filename):
        with open(filename, 'w') as f:
            cg_str = self.to_cg_string()
            f.write(cg_str)

    def get_bulge_angle_stats_core(self, elem, forward=True):
        '''
        Return the angle stats for a particular bulge. These stats describe the
        relative orientation of the two stems that it connects.

        :param elem: The name of the bulge.
        :param connections: The two stems that are connected by it.
        :return: ftms.AngleStat object
        '''
        connections = self.connections(elem)
        if not forward:
            connections = connections[::-1]

        log.debug("elem %s connections %s", elem, connections)
        (stem1, twist1, stem2, twist2, bulge) = ftug.get_stem_twist_and_bulge_vecs(
            self, elem, connections)
        log.debug("stem1 %s, twist1 %s, stem2 %s, twist2 %s, bulge %s",
                  stem1, twist1, stem2, twist2, bulge)

        if round(np.dot(stem1, twist1), 10) != 0 or round(np.dot(stem2, twist2), 10) != 0:
            err = CgIntegrityError("The twists are inconsistent. "
                                   "They should be orthogonal to the corresponding stem vectors."
                                   "Inconsistency found for {},{}".format(elem, connections))
            with log_to_exception(log, err):
                log.error("Angle stem1-twist1 %s dot_product=%s, Angle stem2-twist2 %s degrees dot_product=%s",
                          math.degrees(ftuv.vec_angle(stem1, twist1)
                                       ), np.dot(stem1, twist1),
                          math.degrees(ftuv.vec_angle(stem2, twist2)), np.dot(stem2, twist2),)
            raise err

        try:
            u, v, t, r1, u1, v1 = ftug.get_angle_stat_geometry(
                stem1, twist1, stem2, twist2, bulge)
        except ZeroDivisionError as e:
            with log_to_exception(log, e):
                log.error("Error getting stats for elem %s", elem)
            raise
        dims = self.get_bulge_dimensions(elem)
        ang_type = self.connection_type(elem, connections)
        seq = "&".join(self.get_define_seq_str(elem, adjacent=True))
        log.debug("u %s, v %s", u, v)

        if elem[0] == "m":
            mls = self.find_mlonly_multiloops()
            for ml in mls:
                if elem in ml:
                    descr = self.describe_multiloop(
                        [x for x in ml if x[0] != "s"])
                    if "pseudoknot" in descr:
                        stat_type = "pseudo"
                    elif "open" in descr:
                        stat_type = "open"
                    else:
                        stat_type = "angle"  # ML
                    break
            else:
                assert False
        else:
            stat_type = "angle"  # IL

        angle_stat = ftms.AngleStat(stat_type,
                                    self.name, dims[0], dims[1], u, v, t, r1,
                                    u1, v1, ang_type, self.defines[elem],
                                    seq, self.vposs[elem], self.vbase[elem],
                                    self.vsugar[elem], self.vbackbone[elem])

        return angle_stat

    def get_stats(self, d):
        '''
        Calls get_loop_stat/ get_bulge_angle_stats or get_stem_stats, depending on the element d.

        :returns: A 1- or 2 tuple of stats (2 in case of bulges. One for each direction)
        '''
        if d[0] == "s":
            return (self.get_stem_stats(d),)
        elif d[0] in "mi":
            return self.get_bulge_angle_stats(d)
        else:
            try:
                stat = self.get_loop_stat(d)
            except ValueError:
                if len(self.defines) == 1:
                    # A structure without any stem has no stats.
                    return tuple()
                else:
                    raise
            if d[0] == "f":
                stat.stat_type = "5prime"
            elif d[0] == "t":
                stat.stat_type = "3prime"
            return (stat,)

    def get_loop_stat(self, d):
        '''
        Return the statistics for this loop.

        These stats describe the relative orientation of the loop to the stem
        to which it is attached.

        :param d: The name of the loop
        '''
        loop_stat = ftms.LoopStat()
        loop_stat.pdb_name = self.name

        loop_stat.bp_length = self.get_length(d)
        loop_stat.phys_length = ftuv.magnitude(
            self.coords[d][1] - self.coords[d][0])
        stem1, = self.edges[d]  # Make sure there is only one edge

        (s1b, s1e) = self.get_sides(stem1, d)

        stem1_vec = self.coords[stem1][s1b] - self.coords[stem1][s1e]
        twist1_vec = self.twists[stem1][s1b]
        bulge_vec = self.coords[d][1] - self.coords[d][0]

        # To avoid loops with 0 physical length. (If disconnects in the structure are modelled as loop)
        if ftuv.magnitude(bulge_vec) < 10**-3:
            bulge_vec += (10**-3) * (stem1_vec / ftuv.magnitude(stem1_vec))

        (r, u, v) = ftug.get_stem_separation_parameters(stem1_vec, twist1_vec,
                                                        bulge_vec)
        (loop_stat.r, loop_stat.u, loop_stat.v) = (r, u, v)
        # Will this cause problems in other parts of the code base???
        loop_stat.r = loop_stat.phys_length
        loop_stat.define = self.defines[d]
        loop_stat.seq, = self.get_define_seq_str(d, adjacent=True)
        loop_stat.vres = self.vposs[d]
        loop_stat.vbase = self.vbase[d]
        loop_stat.vsugar = self.vsugar[d]
        loop_stat.vbackbone = self.vbackbone[d]

        if d[0] == "f":
            loop_stat.stat_type = "5prime"
        elif d[0] == "t":
            loop_stat.stat_type = "3prime"
        return loop_stat

    def get_bulge_angle_stats(self, bulge):
        '''
        Return the angle stats for a particular bulge. These stats describe
        the relative orientation of the two stems that it connects.

        :param bulge: The name of the bulge.
        :param connections: The two stems that are connected by it.
        :return: The angle statistics in one direction and angle statistics in
                 the other direction
        '''
        if bulge == 'start':
            return (ftms.AngleStat(), ftms.AngleStat())

        angle_stat1 = self.get_bulge_angle_stats_core(bulge, True)
        angle_stat2 = self.get_bulge_angle_stats_core(bulge, False)
        # If we go into the reverse direction, the first vector of the element basis is inverted.
        # and thus also the last one (cross product of first and second)
        # for k,v in angle_stat2.vres.items():
        #    angle_stat2.vres[k]= v*[-1,1,-1]
        assert round(angle_stat1.get_angle(), 5) == round(angle_stat2.get_angle(
        ), 5), ("{}!={}".format(angle_stat1.get_angle(), angle_stat2.get_angle()))
        return (angle_stat1, angle_stat2)

    def get_stacking_helices(self, method="Tyagi"):
        """
        EXPERIMENTAL

        Return all helices (longer stacking regions) as sets.

        Two stems and one bulge are in a stacking relation, if self.is_stacking(bulge) is true and the stems are connected to the bulge.
        Further more, a stem is in a stacking relation with itself.
        A helix is the transitive closure this stacking relation.

        :returns: A list of sets of element names.
        """
        helices = []
        for d in self.defines:
            if d[0] in "mi" and self.is_stacking(d, method):
                s1, s2 = self.connections(d)
                helices.append(set([d, s1, s2]))
            if d[0] == "s":
                helices.append(set([d]))
        while True:
            for i, j in it.combinations(range(len(helices)), 2):
                stack_bag1 = helices[i]
                stack_bag2 = helices[j]
                if stack_bag1 & stack_bag2:
                    stack_bag1 |= stack_bag2
                    del helices[j]
                    break
            else:
                break
        return helices

    def is_stacking(self, bulge, method="Tyagi", verbose=False):
        """
        EXPERIMENTAL

        Reports, whether the stems connected by the given bulge are coaxially stacking.

        :param bulge: STRING. Name of a interior loop or multiloop (e.g. "m3")
        :param method": STRING. "Tyagi": Use cutoffs from doi:10.1261/rna.305307, PMCID: PMC1894924.
        :returns: A BOOLEAN.
        """

        assert method in ["Tyagi", "CG"]
        if method == "Tyagi":
            return self._is_stacking_tyagi(bulge, verbose)
        return self._is_stacking_CG(bulge, verbose)

    def _is_stacking_CG(self, bulge, verbose=False):
        """"""
        stem1, stem2 = self.connections(bulge)
        angle = ftuv.vec_angle(self.coords[stem1][1] - self.coords[stem1][0],
                               self.coords[stem2][1] - self.coords[stem2][0])
        if angle > math.pi / 2:
            angle = math.pi - angle
        if angle > math.radians(45):
            if verbose:
                print("Angle {}>45".format(math.degrees(angle)))
            return False
        shear_angle1 = ftuv.vec_angle(self.coords[stem1][1] - self.coords[stem1][0],
                                      self.coords[bulge][1] - self.coords[bulge][0])
        if shear_angle1 > math.pi / 2:
            shear_angle1 = math.pi - shear_angle1
        shear_angle2 = ftuv.vec_angle(self.coords[stem2][1] - self.coords[stem2][0],
                                      self.coords[bulge][1] - self.coords[bulge][0])
        if shear_angle2 > math.pi / 2:
            shear_angle2 = math.pi - shear_angle2
        if shear_angle1 > math.radians(60) or shear_angle2 > math.radians(60):
            if verbose:
                print ("Shear angle 1 {}>60 or shear angle 2 {}>60".format(
                    math.degrees(shear_angle1), math.degrees(shear_angle1)))
            return False
        return True

    def _is_stacking_tyagi(self, bulge, verbose=False):
        """
        Implementation of the method described in doi:10.1261/rna.305307 (Tyagi and Matthews) for
        the detection of coaxial stacking.

        Called by self.is_stacking(bulge, "Tyagi")

        ..note::
            This does NOT implement the method for coaxial stacking prediction which is the main
            focus of the paper, only the method for the detection of stacking in pdb files.
        """
        assert bulge[0] in "mi"
        DISTANCE_CUTOFF = [14, 6]
        ANGLE_CUTOFF = [math.acos(0.75), math.acos(0.8)]
        # Relaxed compared to 60 in the paper, because we use
        SHEAR_ANGLE_CUTOFF = math.radians(60)
        # virtual atom positions
        SHEAR_OFFSET_CUTOFF = 10
        if bulge[0] == "m" and self.get_length(bulge) == 0:
            is_flush = True  # flush-stack vs. mismatch-mediated stack
        else:
            is_flush = False

        stem1, stem2 = self.connections(bulge)
        side_nts = self.get_connected_residues(stem1, stem2, bulge)[0]
        # Distance
        try:
            bp_center1 = ftug.get_basepair_center(self, side_nts[0])
            bp_center2 = ftug.get_basepair_center(self, side_nts[1])
        except KeyError:
            return False
        if ftuv.vec_distance(bp_center1, bp_center2) > DISTANCE_CUTOFF[is_flush]:
            if verbose:
                print ("Distance {} > {}".format(ftuv.vec_distance(
                    bp_center1, bp_center2), DISTANCE_CUTOFF[is_flush]))
            return False
        normalvec1 = ftug.get_basepair_plane(self, side_nts[0])
        normalvec2 = ftug.get_basepair_plane(self, side_nts[1])
        # Coaxial
        angle = ftuv.vec_angle(normalvec1, normalvec2)
        if angle > math.pi / 2:
            # Triggered frequently
            #warnings.warn("Angle > 90 degrees: {} ({})".format(angle, math.degrees(angle)))
            angle = math.pi - angle
        if angle > ANGLE_CUTOFF[is_flush]:
            if verbose:
                print ("Angle {} > {}".format(angle, ANGLE_CUTOFF[is_flush]))
            return False
        # Shear Angle
        shear_angle1 = ftuv.vec_angle(normalvec1, bp_center2 - bp_center1)
        if shear_angle1 > math.pi / 2:
            shear_angle1 = math.pi - shear_angle1
        if shear_angle1 > SHEAR_ANGLE_CUTOFF:
            if verbose:
                print ("Shear angle 1 {} > {}".format(
                    shear_angle1, SHEAR_ANGLE_CUTOFF))
            return False
        shear_angle2 = ftuv.vec_angle(normalvec2, bp_center1 - bp_center2)
        if shear_angle2 > math.pi / 2:
            shear_angle2 = math.pi - shear_angle2
        if shear_angle2 > SHEAR_ANGLE_CUTOFF:
            if verbose:
                print ("Shear angle 2 {} > {}".format(
                    shear_angle2, SHEAR_ANGLE_CUTOFF))
            return False
        # Shear Offset
        # Formula for distance between a point and a line
        # from http://onlinemschool.com/math/library/analytic_geometry/p_line/
        if (ftuv.magnitude(np.cross((bp_center1 - bp_center2), normalvec2)) /
                ftuv.magnitude(normalvec2)) > SHEAR_OFFSET_CUTOFF:
            if verbose:
                print ("Shear offset 1 wrong:", (ftuv.magnitude(np.cross((bp_center1 - bp_center2), normalvec2)) /
                                                 ftuv.magnitude(normalvec2)), ">", SHEAR_OFFSET_CUTOFF)
            return False
        if (ftuv.magnitude(np.cross((bp_center1 - bp_center2), normalvec1)) /
                ftuv.magnitude(normalvec1)) > SHEAR_OFFSET_CUTOFF:
            if verbose:
                print ("Shear offset 2 wrong")
            return False
        return True

    def get_stem_stats(self, stem):
        '''
        Calculate the statistics for a stem and return them. These statistics will describe the
        length of the stem as well as how much it twists.

        :param stem: The name of the stem.

        :return: A StemStat structure containing the above information.
        '''
        ss = ftms.StemStat()
        ss.pdb_name = self.name
        #ss.bp_length = abs(self.defines[stem][0] - self.defines[stem][1])
        ss.bp_length = self.stem_length(stem)
        ss.phys_length = ftuv.magnitude(
            self.coords[stem][0] - self.coords[stem][1])
        ss.twist_angle = ftug.get_twist_angle(
            self.coords[stem], self.twists[stem])
        ss.define = self.defines[stem]
        # Stems may or may not have adjacent nucleotides, so we do not include the adj nucleotides
        ss.seqs = self.get_define_seq_str(stem, adjacent=False)
        ss.vbase = self.vbase[stem]
        ss.vsugar = self.vsugar[stem]
        ss.vbackbone = self.vbackbone[stem]
        return ss

    # def get_loop_from_residue(self, residue) ->  use BulgeGraph.get_node_from_residue_num()!
    def _init_coords(self):
        self.coords = LineSegmentStorage(
            self.defines.keys(), on_change=self.reset_vatom_cache)
        self.twists = CoordinateStorage(
            [x for x in self.defines if x[0] == "s"], on_change=self.reset_vatom_cache)

    @property
    def incomplete_elements(self):
        if self._incomplete_elements is None:
            self._incomplete_elements = ftug.get_incomplete_elements(self)
        return set(self._incomplete_elements)

    @property
    def interacting_elements(self):
        if self._interacting_elements is None:
            interacting_nts = []
            for r in self.interacting_residues:
                try:
                    interacting_nts.append(self.seq_id_to_pos(r))
                except ValueError as e:  # interacting missing residue
                    assert "not in list" in str(e)
                    pass
            self._interacting_elements = set(
                self.nucleotides_to_elements(interacting_nts))
        return self._interacting_elements

    def radius_of_gyration(self, method="vres"):
        '''
        Calculate the radius of gyration of this structure.

        :param method: A STRING. one of
                       "fast" (use only coordinates of coarse grained stems) or
                       "vres" (use virtual residue coordinates of stems)

        :return: A number with the radius of gyration of this structure.
        '''
        if len(list(self.stem_iterator())) == 0:
            log.warning(
                "Cannnot calculate ROG (%s) for structure %s without stems. Returning 'nan'", method, self.name)
            return float("nan")
        if method == "fast":
            coords = self.get_ordered_stem_poss()
        elif method == "vres":
            coords = self.get_ordered_virtual_residue_poss()
        else:
            raise ValueError(
                "Wrong method {}. Choose one of 'fast' and 'vres'".format(method))

        rog = ftmd.radius_of_gyration(coords)
        return rog

    def _assign_loop_roles(self, loop):
        """
        EXPERIMENTAL

        For 3-way junctions: Assign the roles P1, P2 and P3
        and J12, J31, J23 according to Lescoute and Westhof,
        RNA 2006 (doi: 10.1261/rna.2208106)

        We try to detect which helices stack and assign the roles in a way
        that J12 is always the stacking segment.

        :returns: None if no clear classification was possible,
                  a dictionary otherwise.
        """
        if len(loop) != 3:
            raise ValueError(
                "junction_family is currentlty only implemented for 3-way junctions.")
        # Find out which helices stack.
        # Version 1: colinearity of helix axes:
        stems = self.edges[loop[0]] | self.edges[loop[1]] | self.edges[loop[2]]
        if len(stems) != 3:
            e = ValueError(
                "The multiloop segments do not form a 3-way junction.")
            with log_to_exception(log, e):
                log.error("stems are %s", stems)
            raise e
        collinearities = {}
        for stem1, stem2 in it.combinations(stems, 2):
            c = ftuv.line_segments_collinearity(
                self.coords[stem1], self.coords[stem2])
            collinearities[(stem1, stem2)] = c
        # The two helices that are most likely to stack.
        stacking_stems = max(collinearities, key=lambda k: collinearities[k])
        if collinearities[stacking_stems] < 0.8:
            return None
        # The second best pair
        second_best_pair, second_best_score = list(
            sorted(collinearities.items(), key=lambda x: x[1]))[1]
        if collinearities[stacking_stems] - second_best_score < 0.5:
            # Two are possible. We look now which one has fewer offset to the 3rd stem.
            # The reference stem could stack with both.
            reference = set(stacking_stems) & set(second_best_pair)
            other = set(stems) - reference
            reference, = reference
            stem1, stem2 = other
            o1 = self.stem_offset(reference, stem1)
            o2 = self.stem_offset(reference, stem2)
            if abs(o1 - o2) < 0.2:  # Angstrom
                log.warning("Cannot assign loop roles. Stacking is ambiguous.")
                return None
            elif o1 < o2:
                stacking_stems = (reference, stem1)
            else:
                stacking_stems = (reference, stem2)
        # We have clearly identified the stacking helices.
        # j12 is the ML-segment that connects the stacking helices.
        j12, = self.edges[stacking_stems[0]] & self.edges[stacking_stems[1]]
        # stem P1 is at the 5' side of the connecting multiloop...
        p1 = self.get_node_from_residue_num(self.define_a(j12)[0])
        # ...and P2 at the 3' side
        p2 = self.get_node_from_residue_num(self.define_a(j12)[1])
        assert set([p1, p2]) == set(stacking_stems)
        assert p1 != p2
        # p3 is the third stem,
        p3, = stems - set([p1, p2])
        # j31 and j23 are the remaining ml segments
        j31, = self.edges[p3] & self.edges[p1]
        j23, = self.edges[p2] & self.edges[p3]
        return {"P1": p1, "P2": p2, "P3": p3, "J12": j12, "J31": j31, "J23": j23}

    def _junction_family_westhof1(self, loop):
        """
        EXPERIMENTAL

        For 3-way junctions: Return the junction family according to
        Lescoute and Westhof, RNA 2006 (doi: 10.1261/rna.2208106).

        In this method, the junction family is defined via the relative
        lengths of fragments, not the orientation of the elements.

        :param loop: Either a dictionary, as returned by self._assign_loop_roles
                     or a list of 3 loop segments.
        """
        if isinstance(loop, dict):
            j_roles = loop
        else:
            j_roles = self._assign_loop_roles(loop)
        # Now compare the lengths of J31 and J23 to determine the junction family
        if self.get_length(j_roles["J31"]) > self.get_length(j_roles["J23"]):
            return "A"
        elif self.get_length(j_roles["J31"]) == self.get_length(j_roles["J23"]):
            return "B"
        else:
            return "C"

    def _junction_family_3d(self, loop):
        """
        EXPERIMENTAL

        For 3-way junctions: Return the junction family depending
        on the 3D orientation.

        :param loop: Either a dictionary, as returned by self._assign_loop_roles
                     or a list of 3 loop segments.
        """
        if isinstance(loop, dict):
            j_roles = loop
        else:
            j_roles = self._assign_loop_roles(loop)
        a31 = self.stem_angle(j_roles["P3"], j_roles["P1"])
        a23 = self.stem_angle(j_roles["P2"], j_roles["P3"])
        # If the stack is perfectly straight, a31+a23=180 degrees.
        if a31 > a23:
            return 2
        else:
            return 1

    def _junction_family_is_perpenticular(self, loop):
        """
        EXPERIMENTAL

        For 3-way junctions Return whether or not the not-stacking loop
        is very roughly perpenticular to the stack.

        :param loop: Either a dictionary, as returned by self._assign_loop_roles
                     or a list of 3 loop segments.
        """
        if isinstance(loop, dict):
            j_roles = loop
        else:
            j_roles = self._assign_loop_roles(loop)
        a31 = self.stem_angle(j_roles["P3"], j_roles["P1"])
        a23 = self.stem_angle(j_roles["P2"], j_roles["P3"])
        # If the stack is perfectly straight, a31+a23=180 degrees.
        if math.radians(60) < a31 < math.radians(120) or math.radians(60) < a23 < math.radians(120):
            return 1
        else:
            return 0

    def stem_offset(self, ref_stem, stem2):
        """
        How much is the offset between the start of stem 2
        and the axis of stem1.

        Assumes that stem1 and stem 2 are connected by a single bulge.
        Then the start of stem2 is defined to be the stem side
        closer to the bulge.
        """
        common_edges = self.edges[ref_stem] & self.edges[stem2]
        if len(common_edges) != 1:
            raise ValueError("Stem1 and stem2 must be connected by a single bulge "
                             "to calculate stem_offset.")
        bulge, = common_edges
        side = self.get_sides(stem2, bulge)[0]
        return ftuv.point_line_distance(self.coords[stem2][side],
                                        self.coords[ref_stem][0],
                                        self.coords.get_direction(ref_stem))

    def stem_angle(self, stem1, stem2):
        """
        Returns the angle between two stems.

        If they are connected via a single element,
        use the direction pointing away from this element for both stems.
        Otherwise, use the direction from start to end.
        """
        vec1 = self.coords.get_direction(stem1)
        vec2 = self.coords.get_direction(stem2)

        common_edges = self.edges[stem1] & self.edges[stem2]
        if len(common_edges) == 1:
            bulge, = common_edges
            if self.get_sides(stem1, bulge) == (1, 0):
                vec1 = -vec1
            else:
                assert self.get_sides(stem1, bulge) == (0, 1)
            if self.get_sides(stem2, bulge) == (1, 0):
                vec2 = -vec2
            else:
                assert self.get_sides(stem2, bulge) == (0, 1)
        return ftuv.vec_angle(vec1, vec2)

    def get_coordinates_array(self):
        '''
        Get all of the coordinates in one large array.

        The coordinates are sorted in the order of the keys
        in coordinates dictionary.

        :return: A 2D numpy array containing all coordinates
        '''
        all_coords = []
        assert len(self.coords) == len(
            self.defines), self.coords.keys() ^ self.defines.keys()
        for key in sorted(self.coords.keys()):
            for i in range(len(self.coords[key])):
                all_coords.append(self.coords[key][i])
        return np.array(all_coords)

    def load_coordinates_array(self, coords):
        '''
        Read in an array of coordinates (as may be produced by get_coordinates_array)
        and replace the coordinates of this structure with it.

        :param coords: A 2D array of coordinates
        :return: self
        '''

        for j, key in enumerate(sorted(self.coords.keys())):
            self.coords[key] = coords[2 * j], coords[2 * j + 1]
        return self

    def get_twists(self, node):
        '''
        Get the array of twists for this node. If the node is a stem,
        then the twists will simply those stored in the array.
        If the node is an interior loop or a junction segment,
        then the twists will be the ones that are adjacent to it,
        projected to the plane normal to the element vector.
        If the node is a hairpin loop or a free end, then the same twist
        will be duplicated and returned twice.

        :param node: The name of the node
        '''
        if node[0] == 's':
            return self.twists[node]

        connections = list(self.edges[node])
        if not connections:
            assert node[0] == "f"
            log.warning("Twists for structures without any stems are 'nan'")
            return np.zeros(3) * float("nan"), np.zeros(3) * float("nan")
        (s1b, s1e) = self.get_sides(connections[0], node)

        if len(connections) == 1:
            vec = ftuv.normalize(ftuv.vector_rejection(
                self.twists[connections[0]][s1b],
                self.coords[node][1] -
                self.coords[node][0]))

            return (vec, vec)

        if len(connections) == 2:
            # interior loop or junction segment
            (s2b, s2e) = self.get_sides(connections[1], node)
            bulge_vec = (self.coords[connections[0]][s1b] -
                         self.coords[connections[1]][s2b])
            return (ftuv.normalize(ftuv.vector_rejection(
                    self.twists[connections[0]][s1b], bulge_vec)),
                    ftuv.normalize(ftuv.vector_rejection(self.twists[connections[1]][s2b], bulge_vec)))

        # uh oh, this shouldn't happen since every node
        # should have either one or two edges
        assert False

    def element_physical_distance(self, element1, element2):
        '''
        Calculate the physical distance between two coarse grain elements.

        :param element1: The name of the first element (e.g. 's1')
        :param element2: The name of the first element (e.g. 's2')
        :return: The closest distance between the two elements.
        '''
        (i1, i2) = ftuv.line_segment_distance(self.coords[element1][0],
                                              self.coords[element1][1],
                                              self.coords[element2][0],
                                              self.coords[element2][1])

        return ftuv.vec_distance(i1, i2)

    def longrange_iterator(self, filter_connected=False):
        '''
        Iterate over all long range interactions in this molecule.

        :param filter_connected: Filter interactions that are between elements
                                 which are connected (mostly meaning multiloops
                                 which connect to the same end of the same stem)
        :return: A generator yielding long-range interaction tuples (i.e. ('s7', 'i2'))
        '''
        seen = set()

        for partner1 in self.longrange.keys():
            for partner2 in self.longrange[partner1]:
                if filter_connected:
                    if self.connected(partner1, partner2):
                        continue

                interaction = tuple(sorted([partner1, partner2]))

                # check if we've already seen this interaction
                if interaction in seen:
                    continue

                seen.add(interaction)

                yield interaction

    def total_length(self):
        '''
        Calculate the combined length of all the elements.
        '''
        total_length = sum(
            [len(list(self.define_residue_num_iterator(d))) for d in self.defines])
        assert total_length == self.seq_length
        return self.seq_length

    def sorted_edges_for_mst(self):
        """
        Keep track of all linked nodes. Used for the generation of the minimal spanning tree.

        This overrides the function in bulge graph and adds an additional sorting criterion
        with lowest priority.
        Elements that have no entry in self.sampled should be preferedly broken.
        This should ensure that the minimal spanning tree is the same after saving
        and loading an RNA to/from a file, if changes of the minimal spanning tree
        were performed by ernwin.
        """
        priority = {'s': 1, 'i': 2, 'm': 3, 'f': 4, 't': 5}
        edges = sorted(it.chain(self.mloop_iterator(),
                                self.iloop_iterator()),
                       key=lambda x: (priority[x[0]], min(self.get_node_dimensions(x)), not x in self.sampled, x))
        return edges

    def coords_to_directions(self):
        """
        The directions of each coarse grain element. One line per cg-element.

        The array is sorted by the corresponding element names alphabetically (`sorted(defines.keys()`)
        The directions usually point away from the elemnt's lowest nucleotide.
        However h,t and f elements always point away from the connected stem.
        """
        coords = self.get_coordinates_array()
        directions = coords[1::2] - coords[0::2]
        return directions

    def coords_from_directions(self, directions):
        """
        Generate coordinates from direction vectors (using also their lengths)

        Currently ignores the twists!

        :param directions: An array of vectors from the side of a cg-element with lower nucleotide number to the side with higher number
                           The array is sorted by the corresponding element names alphabetically (`sorted(defines.keys()`)

        """
        sorted_defines = sorted(self.defines.keys())
        assert len(sorted_defines) == len(directions), "{} != {}".format(
            len(sorted_defines), len(directions))
        if self.build_order is None:
            self.traverse_graph()
        self.coords["s0"] = np.array(
            [0, 0, 0]), directions[sorted_defines.index("s0")]

        for stem1, link, stem2 in self.build_order:  # Bulges and stems
            conn = self.connection_ends(
                self.connection_type(link, [stem1, stem2]))
            link_dir = self.get_link_direction(stem1, stem2, link)
            if link_dir == 1:
                self.coords[link] = self.coords[stem1][conn[0]
                                                       ], self.coords[stem1][conn[0]] + directions[sorted_defines.index(link)]
                if conn[1] == 0:
                    self.coords[stem2] = self.coords[link][1], self.coords[link][1] + \
                        directions[sorted_defines.index(stem2)]
                else:
                    self.coords[stem2] = self.coords[link][1] - \
                        directions[sorted_defines.index(
                            stem2)], self.coords[link][1]
            else:
                self.coords[link] = self.coords[stem1][conn[0]] - \
                    directions[sorted_defines.index(
                        link)], self.coords[stem1][conn[0]]
                if conn[1] == 0:
                    self.coords[stem2] = self.coords[link][0], self.coords[link][0] + \
                        directions[sorted_defines.index(stem2)]
                else:
                    self.coords[stem2] = self.coords[link][0] - \
                        directions[sorted_defines.index(
                            stem2)], self.coords[link][0]
        for d in self.defines:
            if d[0] == "m" and d not in self.mst:
                edges = list(self.edges[d])
                (s1b, _) = self.get_sides(edges[0], d)
                (s2b, _) = self.get_sides(edges[1], d)
                mids1 = self.coords[edges[0]]
                mids2 = self.coords[edges[1]]
                # Save coordinates in direction of the strand.
                if self.get_link_direction(edges[0], edges[1], d) == 1:
                    self.coords[d] = (mids1[s1b], mids2[s2b])
                else:
                    self.coords[d] = (mids2[s2b], mids1[s1b])

            if d[0] in "hft":  # Loops
                stem, = self.edges[d]
                (s1b, _) = self.get_sides(stem, d)
                self.coords[d] = self.coords[stem][s1b], self.coords[stem][s1b] + \
                    directions[sorted_defines.index(d)]

    def virtual_atoms(self, key):
        """
        Get virtual atoms for a key.

        :param key: An INTEGER: The number of the base in the RNA.

        :returns: A dict {atom:coords}, e.g. {"C8":np.array([x,y,z]), ...}
        """
        if isinstance(key, int):
            if key not in self._virtual_atom_cache:
                try:
                    self._virtual_atom_cache[key] = ftug.virtual_atoms(self)[
                        key]
                except KeyError as e:
                    log.info(
                        "Key {} not present. Need to recreate all virtual residues".format(e))
                    self.add_all_virtual_residues()
                    self._virtual_atom_cache[key] = ftug.virtual_atoms(self)[
                        key]
            return self._virtual_atom_cache[key]
        else:
            raise ValueError("Expected an int, found {}".format(key))

    def reset_vatom_cache(self, key):
        """
        Delete all cached information about virtual residues and virtual atoms.
        Used as on_call function for the observing of the self.coords dictionary.

        :param key: A coarse grain element name, e.g. "s1" or "m15"
        """
        try:
            if not self._virtual_atom_cache:
                return
        except AttributeError:  # Happens during deepcopy
            return

        define = self.defines[key]

        # Do not delete self.vposs, it's in the element's coordinate system

        # Delete virtual residues
        try:
            del self.vbases[key]
        except KeyError:
            pass
        try:
            del self.vvecs[key]
        except KeyError:
            pass
        try:
            del self.v3dposs[key]
        except KeyError:
            pass
        try:
            del self.vinvs[key]
        except KeyError:
            pass

        # Delete virtual atoms
        if len(define) > 1:
            for i in range(define[0], define[1] + 1):
                if i in self._virtual_atom_cache:
                    del self._virtual_atom_cache[i]
        if len(define) > 3:
            for i in range(define[2], define[3] + 1):
                if i in self._virtual_atom_cache:
                    del self._virtual_atom_cache[i]
    # def __deepcopy__(self, memo):

    def rotate(self, angle, axis="x", unit="radians"):
        if unit == "degrees":
            angle = math.radians(angle)
        elif unit != "radians":
            raise ValueError(
                "Unit {} not understood. Use 'degrees' or 'radians'".format(unit))
        s = math.sin(angle)
        cosi = math.cos(angle)
        rotation_matrix = ftuv.rotation_matrix(axis, angle)
        self.coords.rotate(rotation_matrix)
        self.twists.rotate(rotation_matrix)
        self.after_coordinates_changed()
        for chain in self.chains.values():
            chain.transform(rotation_matrix.T, [0, 0, 0])

    def rotate_translate(self, offset, rotation_matrix):
        """
        First translate the RNA by offset, then rotate by rotation matrix
        """
        self.coords._coordinates -= offset
        self.coords.rotate(rotation_matrix)
        self.twists.rotate(rotation_matrix)
        self.after_coordinates_changed()
        for chain in self.chains.values():
            chain.transform([[1, 0, 0], [0, 1, 0], [0, 0, 1]], -offset)
            chain.transform(rotation_matrix.T, [0, 0, 0])

    def after_coordinates_changed(self):
        # vposs is in the element coordinate system ==> does not have to be changed.
        #self.vposs = c.defaultdict( dict )
        # Caching for virtual residues
        self.vbases = c.defaultdict(dict)
        self.vvecs = c.defaultdict(dict)
        self.v3dposs= c.defaultdict(dict)
        self.vinvs = c.defaultdict(dict)



#!/usr/bin/env python
# output.py

"""
Author: Sandra Smit (Sandra.Smit@colorado.edu)

Revision History:
File created on 25 Sep 2007.

"""
from __future__ import division
from __future__ import absolute_import
from builtins import zip
from builtins import str
from builtins import map
from .rna2d import Pairs


def ct_output(seq, pairs, header_lines=None, removed=None):
    """Return sequence and structure information in ct format

    seq -- string of sequence characters, no validation
    pairs -- Pairs object or list of tuples
    header_lines -- list of header lines at the top of the file

    Return value is a formatted string
    """
    pairs = Pairs(pairs)
    result = []

    if not header_lines:  # None or empty
        header = '%s ENERGY = 0.0' % (len(seq))
    else:  # take first line. If ct leave as is, otherwise add energy
        head = header_lines[0]
        if 'ENERGY' in head or 'Energy' in head or 'dG' in head:
            header = head.strip()
        else:
            header = '%s ENERGY = 0.0 %s' % (len(seq), head.strip())

    result.append(header)
    partners = pairs.toPartners(len(seq))
    for idx, (seq_symbol, partner_idx) in enumerate(zip(seq, partners)):
        if partner_idx is None:
            partner_idx = -1
        result.append('\t'.join(map(str, [idx + 1, seq_symbol, idx,
                                          idx + 2, partner_idx + 1, idx + 1])))

    if removed is not None:
        result.append('# REMOVED PAIRS FOR: %s' % (header))
        partners = removed.toPartners(len(seq))
        for idx, (seq_symbol, partner_idx) in enumerate(zip(seq, partners)):
            if partner_idx is None:
                partner_idx = -1
            result.append('\t'.join(map(str, [idx + 1, seq_symbol, idx,
                                              idx + 2, partner_idx + 1, idx + 1])))

    return '\n'.join(result)


def bpseq_output(seq, pairs, header_lines=None, removed=None):
    """Return sequence and structure information in bpseq format

    seq -- string of sequence characters, no validation
    pairs -- Pairs object or list of tuples
    header_lines -- list of header lines at the top of the file

    Return value is a formatted string
    """
    result = []
    if header_lines:
        result.extend(header_lines)
    partners = pairs.toPartners(len(seq))
    for idx, (seq_symbol, partner_idx) in enumerate(zip(seq, partners)):
        if partner_idx is None:
            partner_idx = -1
        result.append(
            ' '.join(map(str, [idx + 1, seq_symbol, partner_idx + 1])))

    if removed is not None:
        result.append('# REMOVED BASE PAIRS')
        partners = removed.toPartners(len(seq))
        for idx, (seq_symbol, partner_idx) in enumerate(zip(seq, partners)):
            if partner_idx is None:
                partner_idx = -1
            result.append(
                ' '.join(map(str, [idx + 1, seq_symbol, partner_idx + 1])))

    return '\n'.join(result)


def vienna_output(seq, pairs, header_lines=None, removed=None):
    """Return sequence and structure information in vienna format

    seq -- string of sequence characters, no validation
    pairs -- Pairs object or list of tuples
    header_lines -- list of header lines at the top of the file

    Return value is a formatted string:
    > header information
    sequence
    vienna structure
    """
    result = []
    if not header_lines:
        head = '> Nested structure'
    else:  # take first
        head = '> ' + header_lines[0].strip()
    result.append(head)
    result.append(seq)
    result.append(pairs.toVienna(len(seq)))
    if removed is not None:
        result.append('>REMOVED BASE PAIRS FOR: %s' % (head))
        try:
            result.append(removed.toVienna(len(seq)))
        except:
            result.append(str(removed))
    return '\n'.join(result)


if __name__ == "__main__":
    pass



from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import object
from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

from pprint import pprint

import json
import warnings
import sys
from collections import Counter, defaultdict, namedtuple
import itertools as it
import logging

from logging_exceptions import log_to_exception

from forgi.graph.bulge_graph import RESID

log = logging.getLogger(__name__)


class DSSRLookupError(LookupError):
    pass


class WrongChain(LookupError):
    pass


def dssr_to_pdb_resid(dssr_resid):
    if ":" in dssr_resid:
        model, dssr_resid = dssr_resid.split(":")
        if model != "1":
            return None
    if '.' in dssr_resid:
        chain, _, resid = dssr_resid.partition(".")
    else:
        chain = None
        resid = dssr_resid
    if "^" in resid:
        resid, _, letter = resid.partition("^")
    else:
        letter = " "
    nat = []
    for c in reversed(resid):
        if c.isdigit() or c == "-":
            nat.append(c)
        else:
            break
    resid = "".join(reversed(nat))
    return RESID(chain, (" ", int(resid), letter))


class DSSRAnnotation(object):
    def __init__(self, dssr, cg):
        """
        Store annotation from dssr associated with a coarse grained RNA.

        For DSSR, see: doi: 10.1093/nar/gkv716

        .. note::
            This class is currently compatible with v1.5.7-2016jun16 of dssr.

        :param dssr: A *.json file returned by `x3dna-dssr --json`, or a string holding the json
                     or a dictionary containing the parsed json.
        :param cg: The CoarseGrainedRNA that corresponds to
                   the same pdb file x3dna-dssr was run on.
        """
        if isinstance(dssr, type("")):
            if dssr[0] != "{":
                with open(dssr) as f:
                    dssr = f.read()
            dssr = json.loads(dssr)
        elif not isinstance(dssr, dict):
            raise TypeError(
                "dssr must be an string- or dict instance, not {}".format(type(dssr)))
        self._dssr = dssr
        self._cg = cg

    def aminor_interactions(self):
        if "Aminors" not in self._dssr:
            return
        for aminor in self._dssr["Aminors"]:
            nt1,_,nt23 = aminor["desc_long"].split()
            nt2, nt3 = map(dssr_to_pdb_resid, nt23.split(","))
            nt1 = dssr_to_pdb_resid(nt1)
            if nt1 not in self._cg.seq or nt2 not in self._cg.seq or nt3 not in self._cg.seq:
                log.debug("Ignoring A-Minor interaction outside this connected component: %s->%s-%s", nt1, nt2, nt3)
                continue
            stem = self._cg.get_elem(nt2)
            if stem!=self._cg.get_elem(nt3):
                log.warning("DSSR assumes {} and {} are in a stem, but we assume they are not. Ignoring this AMinor annotation".format(nt2, nt3))
                continue
            yield nt1, stem


    def noncanonical_pairs(self, elem=None):
        """
        Yield all nucleotides which are in a non-canonical basepair
        or a canonical basepair outside a stem.

        This ignores interactions with other RNA chains that are not
        part of the CoarseGrainRNA. Thus, yielded nucleotides are valid indices
        to the CoarseGrainRNA's `.seq` object.

        :param elem: An element name, e.g. "i0".
                     If given, only find basepairs within this secondary structure element.
        :param yields: A triple: `nt1, nt2, basepair-type`
                       nt1 and nt2 are forgi.graph.residue.RESID instances (PDB numbering scheme).
                       The basepair-type is given in the LW notation,
                       e.g. cWH for cis basepairing between the Watson-Crick anfd the Hoogsteen edge.
                       or   tHS for trans basepairing between the Hoogsteen and the Sugar edge.
        """
        for pair in self._dssr["pairs"]:
            nt1 = dssr_to_pdb_resid(pair["nt1"])
            nt2 = dssr_to_pdb_resid(pair["nt2"])
            if nt1 is None or nt2 is None:
                continue
            bp_type = pair["LW"]
            try:
                s1 = self._cg.seq.with_missing[nt1]
                s2 = self._cg.seq.with_missing[nt2]
                log.debug("Basepair is %s -%s- %s", s1, bp_type, s2)
            except IndexError:
                continue
            if elem is None:
                if bp_type == "cWW":
                    partner = self._cg.pairing_partner(nt2)
                    if nt1 == partner:
                        continue
                    elif self._cg.get_node_from_residue_num(nt1)[0] == "s" and self._cg.get_node_from_residue_num(nt2)[0] == "s":
                        log.warning("Pairing partner of %s is %s, but "
                                    "DSSR has %s in interaction. Base-triple?", nt2, partner, nt1)
                yield nt1, nt2, bp_type
            else:
                try:
                    if self._cg.get_node_from_residue_num(nt1) == elem and self._cg.get_node_from_residue_num(nt2) == elem:
                        yield nt1, nt2, bp_type
                except ValueError:
                    pass

    def stacking_nts(self):
        """
        What nucleotides stack according to DSSR

        :returns: A list of tuples of stacking nts.
        """
        if "stacks" not in self._dssr:
            return []
        out = []
        for stack in self._dssr["stacks"]:
            nts = stack["nts_long"].split(",")
            resids = []
            for nt in nts:
                resid = dssr_to_pdb_resid(nt)
                if resid is None:
                    continue
                resids.append(resid)
            for i, resid in enumerate(resids):
                try:
                    out.append((resid, resids[i+1]))
                except IndexError:
                    pass
        return out

    def coaxial_stacks(self):
        """
        Get the element name in the CoarseGrainRNA for all coaxial stacks found by DSSR.
        """
        if "coaxStacks" not in self._dssr:
            return []
        cg_stacks = []
        for dssr_stack in self._dssr["coaxStacks"]:
            cg_stack = []
            try:
                for dssr_stem in dssr_stack["stem_indices"]:
                    cg_stack.append(self.cg_stem(dssr_stem))
            except WrongChain:
                continue
            except RuntimeError:
                continue
            else:
                log.debug("Coaxial stack %s mapped to forgi %s", dssr_stack, cg_stack)
                cg_stacks.append(cg_stack)
        log.debug("coaxial stacks are %s", cg_stacks)
        return cg_stacks

    def cg_stem(self, dssr_stem):
        """
        Get the stem define in the CoarseGrainRNA that corresponds to the stem id in the dssr-format.

        :param dssr_stem: INT the stem in the DSSR Anntotation.
        """
        log.debug("Mapping DSSR stem %s to forgi", dssr_stem)
        if "stems" not in self._dssr:
            raise DSSRLookupError("The DSSR object does not contain any stem!")
        for stem_obj in self._dssr["stems"]:
            if stem_obj["index"] == dssr_stem:
                break
        else:
            raise DSSRLookupError("No stem with index {}".format(dssr_stem))
        log.debug("Found stem %s&%s", stem_obj["strand1"], stem_obj["strand2"])

        cg_stems = Counter()  # See, if the dssr_stems maps to more than 1 cg-stem
        for pair in stem_obj["pairs"]:
            res1 = dssr_to_pdb_resid(pair["nt1"])
            res2 = dssr_to_pdb_resid(pair["nt2"])
            if res1 is None or res2 is None:
                continue
            log.debug("Contains pair %s-%s", res1, res2)
            if self._cg.chains and (res1.chain not in self._cg.chains or res2.chain not in self._cg.chains):
                e = WrongChain()
                with log_to_exception(log, e):
                    log.error("Wrong chain: res1={}, res2={}, cg.chains={}".format(
                        res1, res2, self._cg.chains))
                raise e
            i1 = self._cg.seq.to_integer(res1)
            i2 = self._cg.seq.to_integer(res2)
            nodes = self._cg.nucleotides_to_elements([i1, i2])
            for node in nodes:
                cg_stems[node] += 1
        if not cg_stems:
            raise RuntimeError(
                "No stem matching dssr_stem {}.".format(dssr_stem))
        most_common = cg_stems.most_common()
        if len(most_common) > 1:
            extra_info = ""
            for d in cg_stems.keys():
                if d[0] == "i":
                    extra_info += "\n{} is {}:".format(d,
                                                       self._cg.get_define_seq_str(d))
                    extra_info += "\n\t" + self._cg.seq + "\n\t" + \
                        self._cg.to_dotbracket_string() + "\n\t"
                    resnums = list(self._cg.define_residue_num_iterator(d))
                    for i in range(len(self._cg.seq)):
                        pos = i + 1
                        if pos in resnums:
                            extra_info += "^"
                        else:
                            extra_info += " "

            warnings.warn("dssr_stem {} maps to more than one cg element: {} {}".format(
                dssr_stem, list(cg_stems.keys()), extra_info))
        for mc in most_common:
            if mc[0][0] == "s":
                return mc[0]
        raise RuntimeError("No stem matching dssr_stem {}, only single stranded region: {}.".format(
            dssr_stem, list(cg_stems.keys())))

    def compare_coaxial_stack_annotation(self, forgi_method="Tyagi", allow_single_bp=False):
        """
        Compare the coaxial stack annotation between the DSSR method and the forgi method.
        :param forgi_method: "Tyagi" or "CG". Method for stack detection in forgi.
        """
        stacks_dssr = set()
        stacks_forgi = set()
        Stack = namedtuple('Stack', ['stems', 'forgi', 'dssr'])
        dssr_stacks = self.coaxial_stacks()
        for stack in dssr_stacks:
            if stack[0] == stack[1]:
                stacks_dssr.add(Stack(tuple(stack), "one helix", "stacking"))
            else:
                bulges = (self._cg.edges[stack[0]] & self._cg.edges[stack[1]])
                if not bulges:
                    stacks_dssr.add(
                        Stack(tuple(stack), "not connected", "stacking"))
                else:
                    bulge = bulges.pop()  # We only look at one bulge. If 2 stems are connected by more than 1 bulge, cg.is_stacking should give the same result in both cases. TODO: Write a test for tihis case!
                    if self._cg.is_stacking(bulge, forgi_method):
                        curr_stack = Stack(
                            tuple(stack), "stacking", "stacking")
                        stacks_dssr.add(curr_stack)
                        stacks_forgi.add(curr_stack)
                    else:
                        stacks_dssr.add(
                            Stack(tuple(stack), "not stacking", "stacking"))
        for d in self._cg.defines:
            if d[0] not in "mi":
                continue
            if self._cg.is_stacking(d, forgi_method):
                s1, s2 = self._cg.connections(d)
                if not allow_single_bp and (self._cg.stem_length(s1) == 1 or self._cg.stem_length(s2) == 1):
                    continue
                if [s1, s2] not in dssr_stacks and [s2, s1] not in dssr_stacks:
                    stacks_forgi.add(
                        Stack((s1, s2), "stacking", "not stacking"))
        return stacks_forgi, stacks_dssr

    def stacking_loops(self):
        """
        Returns a list of loop (multiloop and interior-loop) elements,
        along which the adjacent stems stack.

        Note that not all nucleotides of these elements have to participate in
        the stack and additional nucleotides can be used for the stack.

        I fact, this only checks whether one nucleotide of each of the first
        (seen from the side of the loop element) basepairs of the adjacent
        stems are in a continuouse stack.

        :returns: A list of element names
        """
        stacking = []
        for loop in self._cg.defines:
            if loop[0] not in "im":
                continue
            nts_elem = []
            define_a = self._cg.define_a(loop)
            pos1 = define_a[0]
            helix1 = [self._cg.seq.to_resid(pos1), self._cg.seq.to_resid(
                self._cg.pairing_partner(pos1))]
            pos2 = define_a[1]
            helix2 = [self._cg.seq.to_resid(pos2), self._cg.seq.to_resid(
                self._cg.pairing_partner(pos2))]
            for stack in self._dssr.get("stacks", []):
                nts_dssrstack = []
                for nt in stack["nts_long"].split(","):
                    nt = dssr_to_pdb_resid(nt)
                    if nt is None:
                        continue
                    nts_dssrstack.append(nt)
                if any(nt in nts_dssrstack for nt in helix1) and any(nt in nts_dssrstack for nt in helix2):
                    log.debug("elem %s is stacking ", loop)
                    stacking.append(loop)
                    break
        log.debug(stacking)
        return stacking

    def basepair_stacking(self, forgi_method="Tyagi"):
        # if "_" in self._cg.name:
        #    chainname="chain_"+self._cg.name.split("_")[-1]
        #    forgi_chain=self._cg.name.split("_")[-1]
        # else:
        if True:
            chainname = "all_chains"
            forgi_chain = None
        dssr_helices = []
        for stem in self._dssr.get("stems", []):
            dssr_helices.append(set())
            for pair in stem["pairs"]:
                nt1 = dssr_to_pdb_resid(pair["nt1"])
                nt2 = dssr_to_pdb_resid(pair["nt2"])
                if nt1 is None or nt2 is None:
                    continue
                if forgi_chain is None or nt1[0] == forgi_chain:
                    dssr_helices[-1].add(nt1)
                if forgi_chain is None or nt2[0] == forgi_chain:
                    dssr_helices[-1].add(nt2)
            if not dssr_helices[-1]:
                del dssr_helices[-1]
        for stack in self._dssr.get("stacks", []):
            dssr_helices.append(set())
            for nt in stack["nts_long"].split(","):
                nt1 = dssr_to_pdb_resid(nt)
                if nt1 is not None and forgi_chain is None or nt1[0] == forgi_chain:
                    dssr_helices[-1].add(nt1)
            if not dssr_helices[-1]:
                del dssr_helices[-1]
        # Merge stacks that overlap.
        while True:
            for i, j in it.combinations(range(len(dssr_helices)), 2):
                stack_bag1 = dssr_helices[i]
                stack_bag2 = dssr_helices[j]
                if stack_bag1 & stack_bag2:
                    stack_bag1 |= stack_bag2
                    del dssr_helices[j]
                    break
            else:
                break
        element_helices = self._cg.get_stacking_helices(forgi_method)
        forgi_helices = []
        for helix in element_helices:
            forgi_helices.append(set())
            for d in helix:
                forgi_helices[-1] |= set(
                    n for n in self._cg.define_residue_num_iterator(d, seq_ids=True))
        print("FH", forgi_helices)
        for i, d_helix in enumerate(dssr_helices):
            no_match = True
            for j, f_helix in enumerate(forgi_helices):
                if d_helix & f_helix:
                    no_match = False
                    print ("dssr helix {} matches forgi helix {}. Only forgi: {}nts, only dssr: {}nts, both: {}nts".format(
                        i, j, len(f_helix - d_helix), len(d_helix - f_helix), len(d_helix & f_helix)))
            if no_match:
                print("dssr helix {}: No match ({}nts)".format(i, len(d_helix)))
        for j, f_helix in enumerate(forgi_helices):
            no_match = True
            for i, d_helix in enumerate(dssr_helices):
                if d_helix & f_helix:
                    no_match = False
                    print ("forgi helix {} matches dssr helix {}. Only forgi: {}nts, only dssr: {}nts, both: {}nts".format(
                        j, i, len(f_helix - d_helix), len(d_helix - f_helix), len(d_helix & f_helix)))
            if no_match:
                print("forgi helix {}: No match ({}nts)".format(j, len(f_helix)))
        print("FH", forgi_helices)
        print ("seq   " + self._cg.seq)
        print ("forgi " + self._cg.to_dotbracket_string())
        display_f = []
        for helix in forgi_helices:
            display_f.append([self._cg.seq_id_to_pos(nt) - 1 for nt in helix])
        display_f.sort(key=lambda x: min(x))
        helixstri = "forgi "
        for i in range(len(self._cg.seq)):
            for j, dp in enumerate(display_f):
                if i in dp:
                    # Start with 0 (ord("0")==48 and display only characters between chr(33)="!" and chr(126)="~"
                    chrid = j + 15
                    if chrid > 93:
                        chrid = chrid % (93)
                    chrid = chrid + 33

                    helixstri += chr(chrid)
                    break
            else:
                helixstri += " "
        print(helixstri)
        print ("DSSR  " + self._dssr["dbn"][chainname]["sstr"])
        display_d = []
        # print(dssr_helices)
        for helix in dssr_helices:
            try:
                display_d.append(
                    [self._cg.seq_id_to_pos(nt) - 1 for nt in helix])
            except ValueError as e:
                log.exception("Exception occurred")
        display_d.sort(key=lambda x: min(x))
        helixstri = "dssr  "
        for i in range(len(self._cg.seq)):
            for j, dp in enumerate(display_d):
                if i in dp:
                    chrid = j + 15
                    if chrid > 93:
                        chrid = chrid % (93)
                    chrid = chrid + 33

                    helixstri += chr(chrid)
                    break
            else:
                helixstri += " "
        print(helixstri)
        print(dssr_helices)
        dssr_to_elems = [self._cg.nucleotides_to_elements(
            list(map(self._cg.seq.to_integer, helix))) for helix in dssr_helices]
        print(dssr_to_elems)

    def compare_dotbracket(self):
        if "_" in self._cg.name:
            chainname = "chain_" + self._cg.name.split("_")[-1]
        else:
            chainname = "all_chains"
        print ("seq   " + self._cg.seq)
        print ("forgi " + self._cg.to_dotbracket_string())
        print ("DSSR  " + self._dssr["dbn"][chainname]["sstr"])



#!/usr/bin/python

from __future__ import print_function
from __future__ import division
from builtins import str
from builtins import zip
from builtins import map
from builtins import object

import csv
import itertools as it
import sys
import warnings

import random as rand
import numpy as np
import numpy.random as nr
import collections as c
import math as m
import math

import logging
log = logging.getLogger(__name__)

import forgi.utilities.debug as fud
import forgi.threedee.utilities.vector as ftuv
import forgi.threedee.utilities.virtual_residues as ftuvres
import forgi.threedee.utilities.graph_pdb as ftug

# The two constants seem to be unused.
avg_stem_bp_length = 2.24
avg_twist_rotation_per_bp = 360 / 11.


class LoopStat(object):
    '''
    Class for storing the individual statistics about loops.

    phys_length: The length between the start and the centroid of the loop.
    '''

    def __init__(self, line='', s_type="loop"):
        self.stat_type = s_type
        self.pdb_name = ''

        self.bp_length = 0
        self.phys_length = 0.
        self.r = 0.
        self.u = 0.
        self.v = 0.

        self.define = []
        self.seq = ""
        self.vres = {}
        self.vbase = {}
        self.vsugar = {}
        self.vbackbone = {}
        if len(line) > 0:
            try:
                self.parse_line(line)
            except:
                print("Error parsing line:", line, file=sys.stderr)
                raise

    def parse_line(self, line):
        '''
        Parse a line containing statistics about the shape of a stem.

        :param line: The line from the statistics file.
        '''

        parts = line.strip().split()

        self.pdb_name = parts[1]

        self.bp_length = int(parts[2])
        self.phys_length = float(parts[3])
        self.r = float(parts[3])
        self.u = float(parts[4])
        self.v = float(parts[5])
        if len(parts) > 6:
            self.define = list(map(int, [parts[6], parts[7]]))
            self.seq = parts[8]
        if len(parts) > 9:
            self.vres, self.vbase , self.vsugar, self.vbackbone = ftuvres.parse_vres(parts[9:])

    def __str__(self):
        out = ("{stat_type} {pdb_name} {bp_length} {phys_length}"
               " {u} {v} ".format(**self.__dict__))
        out += " ".join(map(str, self.define)) + " " + self.seq
        out += " " + ftuvres.serialize_vres(self.vres)
        out += " vbase " + ftuvres.serialize_vres(self.vbase)
        out += " vsugar " + ftuvres.serialize_vres(self.vsugar)
        out += " vbackbone " + ftuvres.serialize_vres(self.vbackbone)
        return out

    def __eq__(self, other):
        if type(self) == type(other):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        return not self == other


class StemStat(object):
    '''
    Class for storing the individual statistics about helices.

    Each stem will be defined by its base pair length. Two
    parameters are associated with each base pair length:

    phys_length: The physical length of such a helix
    twist_angle: The angle between its two twist segments
    '''

    def __init__(self, line=''):
        self.pdb_name = ''

        self.bp_length = 0
        self.phys_length = 0.

        self.twist_angle = 0.
        self.define = []
        self.seqs = []
        self.vbase = {}
        self.vsugar = {}
        self.vbackbone = {}
        if len(line) > 0:
            self.parse_line(line)


    def parse_line(self, line):
        '''
        Parse a line containing statistics about the shape of a stem.

        :param line: The line from the statistics file.
        '''

        parts = line.strip().split()

        self.pdb_name = parts[1]

        self.bp_length = int(parts[2])
        self.phys_length = float(parts[3])
        self.twist_angle = float(parts[4])
        if len(parts) > 5:
            self.define = [int(parts[5]), int(parts[6]),
                           int(parts[7]), int(parts[8])]
        try:
            self.seqs = [parts[9], parts[10]]
        except IndexError:
            pass
        else:
            _, self.vbase , self.vsugar, self.vbackbone = ftuvres.parse_vres(parts[11:])

    def __str__(self):
        try:
            return ("stem {pdb_name} {bp_length} {phys_length} {twist_angle} ".format(**self.__dict__)
                    + " ".join(map(str, self.define)) + " " + " ".join(self.seqs)
                    + " vbase {} vsugar {} vbackbone {} ".format(
                        ftuvres.serialize_vres(self.vbase),
                        ftuvres.serialize_vres(self.vsugar),
                        ftuvres.serialize_vres(self.vbackbone)))
        except (KeyError, IndexError):
            warnings.warn(
                "Could not print define '{}' for StemStat".format(self.define))
            return "stem {} {} {} {}".format(self.pdb_name, self.bp_length, self.phys_length, self.twist_angle)


    def __eq__(self, other):
        if type(self) == type(other):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        return not self == other


class AngleStat(object):
    '''
    Class for storing an individual statistic about inter-helical angles.
    '''

    def __init__(self, stat_type="angle", pdb_name='', dim1=0, dim2=0, u=0, v=0, t=0, r1=0, u1=0, v1=0, ang_type='x',
                 define=[], seq="", vres={}, vbase={}, vsugar={}, vbackbone={}):
        #log.debug("Stat init called")
        self.pdb_name = pdb_name
        self.dim1 = dim1
        self.dim2 = dim2
        self.stat_type = stat_type

        self.u = u
        self.v = v
        self.t = t

        self.r1 = r1
        self.u1 = u1
        self.v1 = v1

        self.ang_type = ang_type

        self.define = define
        self.seq = seq
        self.vres = vres
        self.vbase = vbase
        self.vsugar = vsugar
        self.vbackbone = vbackbone

    def __eq__(self, a_s):
        '''
        Is this AngleStat equal to another one.
        '''
        log.debug("Comparing angle stats {} and {}".format(
            self.pdb_name, a_s.pdb_name))
        if self.dim1 != a_s.dim1:
            log.debug("Dim1 {} != {}".format(self.dim1, a_s.dim1))
            return False
        if self.dim2 != a_s.dim2:
            log.debug("Dim2 {} != {}".format(self.dim2, a_s.dim2))
            return False
        if not np.allclose(self.u, a_s.u):
            log.debug("u {} != {}".format(self.u, a_s.u))
            return False
        if not np.allclose(self.v, a_s.v):
            log.debug("v {} != {}".format(self.v, a_s.v))
            return False
        if not np.allclose(self.t, a_s.t):
            log.debug("t {} != {}".format(self.t, a_s.t))
            return False
        if not np.allclose(self.r1, a_s.r1):
            log.debug("r1 {} != {}".format(self.r1, a_s.r1))
            return False
        if not np.allclose(self.u1, a_s.u1):
            log.debug("u1 {} != {}".format(self.u1, a_s.u1))
            return False
        if not np.allclose(self.v1, a_s.v1):
            log.debug("v1 {} != {}".format(self.v1, a_s.v1))
            return False
        log.debug("Angle stats comparing equal")
        return True

    def parse_line(self, line):
        parts = line.strip().split()

        self.stat_type = parts[0]
        self.pdb_name = parts[1]

        self.dim1 = int(parts[2])
        self.dim2 = int(parts[3])

        self.u = float(parts[4])
        self.v = float(parts[5])
        self.t = float(parts[6])

        self.r1 = float(parts[7])
        self.u1 = float(parts[8])
        self.v1 = float(parts[9])

        self.ang_type = int(parts[10])

        if self.dim2 == 1000:
            # multiloop or single stranded region
            if self.dim1 == 0:
                # no unpaired bases means no define
                def_len = 0
            else:
                def_len = 2
        else:
            def_len = 4
            if self.dim1 == 0:
                def_len -= 2
            if self.dim2 == 0:
                def_len -= 2

            # for interior loops, at least one strand has to
            # have an unpaired base
            assert(def_len > 0)

        self.define = list(map(int, parts[11:11 + def_len]))
        self.seq = parts[11 + def_len]
        log.debug("len(parts)=%s, def_len=%s", len(parts), def_len)
        if len(parts) > 12 + def_len:
            self.vres, self.vbase , self.vsugar, self.vbackbone = ftuvres.parse_vres(parts[12 + def_len:])

    def orientation_params(self):
        '''
        Return a tuple containing the parameters which specify the orientation
        of one stem with respect to another.

        :return: (u, v)
        '''

        return (self.u, self.v)

    def twist_params(self):
        '''
        Returns a tuple containing the parameters which specify the difference in
        twist between two stems.

        :return (u, v, t)
        '''
        return (self.u, self.v, self.t)

    def position_params(self):
        '''
        Return a tuple containing the parameters which specify the position
        of one stem with respect to another.

        :return: `(r1, u1, v1)`
        '''
        return (self.r1, self.u1, self.v1)

    def __str__(self):
        out_str = "%s %s %d %d %f %f %f %f %f %f %d %s %s %s vbase %s vsugar %s vbackbone %s" % (self.stat_type,
                                                                 self.pdb_name,
                                                                 self.dim1,
                                                                 self.dim2,
                                                                 self.u,
                                                                 self.v,
                                                                 self.t,
                                                                 self.r1,
                                                                 self.u1,
                                                                 self.v1,
                                                                 self.ang_type,
                                                                 " ".join(
                                                                     map(str, self.define)),
                                                                 self.seq,
                                                                 ftuvres.serialize_vres(self.vres),
                                                                 ftuvres.serialize_vres(self.vbase),
                                                                 ftuvres.serialize_vres(self.vsugar),
                                                                 ftuvres.serialize_vres(self.vbackbone))
        return out_str

    def __hash__(self):
        return hash(str(self))

    def get_angle(self):
        '''
        Return the angle between the two connected stems.
        '''
        return ftuv.vec_angle(np.array([-1., 0., 0.]), ftuv.spherical_polar_to_cartesian([1, self.u, self.v]))

    def deviation_from(self, stat2):
        """
        How much does the other stat differ from this stat?

        :param stat2: Another AngleStat
        :returns: A 4-tuple: The positional deviation in Angstrom, and 3 absolute angular deviations in radians.
                  The  angular deviations are u, v and t
        """
        ret = []
        pos1 = ftuv.spherical_polar_to_cartesian(self.position_params())
        pos2 = ftuv.spherical_polar_to_cartesian(stat2.position_params())
        ret.append(ftuv.magnitude(pos1 - pos2))
        log.debug("Position difference is %f", ret[-1])

        for attr in ["u", "v", "t"]:
            raw_diff = getattr(self, attr) - getattr(stat2, attr)
            # Map the difference to a value between 0 and pi
            raw_diff_on_circle = abs(
                (raw_diff + math.pi / 2) % (math.pi) - math.pi / 2)
            log.debug("Angular difference for %s is %f, mapped to %f",
                      attr, raw_diff, raw_diff_on_circle)
            ret.append(raw_diff_on_circle)
        return tuple(ret)

    def is_similar_to(self, stat2, position_cutoff=4, angular_cutoff=None):
        """
        Returns True, if theAngleStat is similar
        (according to the specified cutoff) to the other Angle Stat.

        :param position_cutoff: in angstrom
        :param angular_cutoff: in radians. If not given, uses the position cutoff as a value in degrees
        """
        if angular_cutoff is None:
            angular_cutoff = math.radians(position_cutoff)
        deviation = self.deviation_from(stat2)
        if deviation[0] > position_cutoff:
            log.debug(
                "Dissimilar, because of position deviation = %f", deviation[0])
            return False
        for dev in deviation[1:]:
            if dev > angular_cutoff:
                log.debug("Dissimilar, because of angular deviation %s (%s) > %f",
                          dev, deviation[1:], angular_cutoff)
                return False
        log.debug("%s < (%f, %f)", deviation, position_cutoff, angular_cutoff)
        return True


class RandomAngleStats(object):
    '''
    Store all of the angle stats.
    '''

    def __init__(self, discrete_angle_stats):
        self.cont_stats = dict()
        self.make_random(discrete_angle_stats)

    def create_random_function(self, data):
        '''
        Create a function that returns a random value for
        each column in the statistics.

        :param stats: A table containing n rows and m columns
        :return: A function returning m random values with a
                 maximum and minimum no greater than the largest
                 and least values in that column, respectively.
        '''
        mins = map(min, data.T)
        maxs = map(max, data.T)

        def bounded_uniform():
            return [nr.uniform(i, x) for i, x in zip(mins, maxs)]

        return bounded_uniform

    def make_random(self, discrete_angle_stats):
        '''
        Create a set of statistsics that is random in each direction.

        The maximum and minimum u and v values will be taken
        from the discrete statistics.
        '''
        import scipy.stats as ss
        for key1, key2, key3 in discrete_angle_stats.keys():
            dims = (key1, key2, key3)
            data = []

            for d in discrete_angle_stats[(key1, key2, key3)]:
                data += [[d.u, d.v, d.t, d.r1, d.u1, d.v1]]

            '''
            if len(data) < 3:
                continue
            '''

            try:
                self.cont_stats[dims] = self.create_random_function(
                    np.array(data))
            except np.linalg.LinAlgError as lae:
                print("Singular matrix, dimensions:", dims, file=sys.stderr)

    def sample_stats(self, dims):
        '''
        Sample a set of statistics.

        :param dims: The dimensions of the bulge for which to sample.
        '''
        new_stats = self.cont_stats[dims]()
        s = AngleStat()
        (s.u, s.v, s.v, s.r1, s.u1, s.v1) = new_stats
        return s


class ConstructionStats(object):
    angle_stats = None
    stem_stats = None
    loop_stats = None
    fiveprime_stats = None
    threeprime_stats = None
    conf_stats = None


def get_angle_stats(filename, refresh=False):
    '''
    Load the statistics about inter the helix-helix orientations from a file.

    The file format should be as follows:

    `angle pdb_name dim1 dim2 r u v t r1 u1 v1 s1b s2b`

    Where the parameters are as follows:

    * `angle`: identifier for a type of statistics... should always just be 'angle'
    * `pdb_name`: the name of the pdb file these statistics came from
    * `dim1`: the smaller dimension of the bulge
    * `dim2`: the larger dimension of the bulge
    * `u`: the polar angle of the orientation of the 2nd stem
    * `v`: the azimuth of the orientation of the 2nd stem
    * `t`: the orientation of the twist of the second stem
    * `r1`: the distance of the start of the 2nd stem helix from the end of the 1st
            stem helix
    * `u1`: the polar angle of the separation vector of the two helices
    * `v1`: the azimuth of the separation vector of the two helices
    * `s1b`: The side of the first stem closest to the bulge
    * `s2b`: The side of the second stem closest to the bulge

    The azimuth is always defined with respect to the coordinate system defined
    by the stem1 helix axis vector and it's twist vector (the one adjacent to the
    bulge element).
    '''
    if ConstructionStats.angle_stats != None and not refresh:
        return ConstructionStats.angle_stats

    '''
    import pickle
    ConstructionStats.angle_stats = pickle.load(open('fess/stats/angle_stats.pickle', 'r'))

    print >>sys.stderr, "done loading stats"
    return ConstructionStats.angle_stats

    '''
    ConstructionStats.angle_stats = c.defaultdict(list)
    # DefaultDict(DefaultDict([]))

    f = open(filename, 'r')

    count = 0
    for line in f:
        line = line.strip()
        if line.startswith('angle'):
            angle_stat = AngleStat()
            angle_stat.parse_line(line)

            if len(angle_stat.define) > 0 and angle_stat.define[0] == 1:
                continue

            ConstructionStats.angle_stats[(
                angle_stat.dim1, angle_stat.dim2, angle_stat.ang_type)] += [angle_stat]
            ConstructionStats.angle_stats[(
                angle_stat.dim2, angle_stat.dim1, -angle_stat.ang_type)] += [angle_stat]
            count += 1

    f.close()

    '''
    pickle.dump(ConstructionStats.angle_stats, open('fess/stats/angle_stats.pickle','w'))
    '''

    #print >>sys.stderr, "done loading stats"
    return ConstructionStats.angle_stats


def get_angle_stat_dims(s1, s2, angle_type, min_entries=1):
    '''
    Return a list of tuples which indicate the dimensions for which angle
    stats are avilable.

    :param s1: The first size
    :param s2: The second size
    :param angle_type: The type of the angle.
    :param min_entries: The minimum number of stats that have to be available
    '''
    available_stats = []
    angle_stats = get_angle_stats()

    for (k1, k2, k3) in angle_stats.keys():
        # and len(angle_stats[(k1,k2,k3)]) >= min_entries: #BT: I think this should't be here.
        if k3 == angle_type:
            dist = m.sqrt((k1 - s1) ** 2 + (k2 - s2) ** 2)
            available_stats += [(dist, (k1, k2, k3))]

    available_stats.sort()
    return available_stats


def get_one_d_stat_dims(d, stats, min_entries=1):
    available_stats = []

    for k in stats.keys():
        available_stats += [(abs(d - k), k)]

    available_stats.sort()
    return available_stats


def get_stem_stats(filename, refresh=False):
    '''
    Load the statistics from the file.

    format:

    stem pdb_name bp_length phys_length twist_angle

    :param filename: The name of the file.
    '''
    if ConstructionStats.stem_stats is not None and not refresh:
        return ConstructionStats.stem_stats

    ConstructionStats.stem_stats = c.defaultdict(list)

    f = open(filename, 'r')
    #print("Opening file", filename)
    for line in f:
        if line.strip().find('stem') == 0:
            stem_stat = StemStat(line)
            '''
            if stem_stat.define[0] == 1:
                continue
            '''

            ConstructionStats.stem_stats[(
                stem_stat.bp_length, stem_stat.bp_length)] += [stem_stat]

    f.close()

    return ConstructionStats.stem_stats


def get_fiveprime_stats(filename, refresh=False):
    '''
    Load the statistics from the file.

    format:

    `fiveprime pdb_name bp_length phys_length`

    :param filename: The name of the file.
    '''
    if ConstructionStats.fiveprime_stats is not None and not refresh:
        return ConstructionStats.fiveprime_stats

    ConstructionStats.fiveprime_stats = c.defaultdict(list)

    f = open(filename, 'r')

    for line in f:
        if line.strip().find('5prime') == 0:
            fiveprime_stat = LoopStat(line)
            ConstructionStats.fiveprime_stats[fiveprime_stat.bp_length] += [
                fiveprime_stat]

    f.close()

    return ConstructionStats.fiveprime_stats


def get_threeprime_stats(filename, refresh=False):
    '''
    Load the statistics from the file.

    format:

    `threeprime pdb_name bp_length phys_length`

    :param filename: The name of the file.
    '''
    if ConstructionStats.threeprime_stats is not None and not refresh:
        return ConstructionStats.threeprime_stats

    ConstructionStats.threeprime_stats = c.defaultdict(list)

    f = open(filename, 'r')

    for line in f:
        if line.strip().find('3prime') == 0:
            threeprime_stat = LoopStat(line)
            ConstructionStats.threeprime_stats[threeprime_stat.bp_length] += [
                threeprime_stat]

    f.close()

    return ConstructionStats.threeprime_stats


def get_loop_stats(filename, refresh=False):
    '''
    Load the statistics from the file.

    format:

    `loop pdb_name bp_length phys_length`

    :param filename: The name of the file.
    '''
    if ConstructionStats.loop_stats != None and not refresh:
        return ConstructionStats.loop_stats

    ConstructionStats.loop_stats = c.defaultdict(list)

    f = open(filename, 'r')

    for line in f:
        if line.strip().find('loop') == 0:
            loop_stat = LoopStat(line)
            ConstructionStats.loop_stats[loop_stat.bp_length] += [loop_stat]

    f.close()

    return ConstructionStats.loop_stats


class ClusteredAngleStats(object):
    def __init__(self, filename):
        """
        A collection of clustered angle_stats. Whenever stats for a certain key are retrieved,
        only one (randomly choosen) representative per cluster is returned.

        Requires a clustered angle stats file (created by fess/scripts/cluster_stats.py).
        The file should look like this:

        '
        # Cluster 0 for (1, 1, -1):
        angle RS_1788_S_000008_A 1 1 1.520877 1.837257 -1.380352 11.237569 0.867246 2.057771 1 19 19 30 30 UCG CAA
        # Cluster 1 for (1, 1, -1):
        angle RS_1140_S_000002_A 1 1 1.563667 1.869580 -0.967966 16.241881 0.896602 2.177000 -1 13 13 24 24 ACG CUU
        angle RS_1183_S_000008_A 1 1 1.533551 1.914482 -1.108396 15.780723 0.829850 2.255556 -1 4 4 46 46 UAG CAG
        '

        :param filename: The filename of the clustered angle stats file.
        """
        #: A dict `(dim1, dim2, ang_type)` : list of lists.
        #: Each value is a list of clusters, and each cluster is a list.
        self._stats_dict = c.defaultdict(list)
        lastkey = None
        with open(filename) as f:
            for line in f:
                if line.startswith('# Cluster'):
                    fields = line.split()
                    dim1 = int(fields[4].lstrip("(").rstrip(","))
                    dim2 = int(fields[5].rstrip(","))
                    ang_type = int(fields[6].rstrip("):"))
                    self._stats_dict[(dim1, dim2, ang_type)].append([])
                    lastkey = (dim1, dim2, ang_type)
                elif line.startswith('#'):
                    continue
                else:
                    angle_stat = AngleStat()
                    angle_stat.parse_line(line)
                    # I don't know what this does, I just copied the following 2 lines from
                    # Peter's code in get_angle_stats
                    if len(angle_stat.define) > 0 and angle_stat.define[0] == 1:
                        continue

                    assert lastkey == (angle_stat.dim1, angle_stat.dim2, angle_stat.ang_type) or lastkey == (
                        angle_stat.dim2, angle_stat.dim1, -angle_stat.ang_type)
                    self._stats_dict[lastkey][-1].append(angle_stat)

    def __getitem__(self, key):
        """
        Returns a list of stats with only one randomly choosen stat for each cluster.
        """
        return [rand.choice(x) for x in self._stats_dict[key]]

    def keys(self):
        return self._stats_dict.keys()

    def lookup_stat(self, stat):
        key = (stat.dim1, stat.dim2, stat.ang_type)
        clusters = self._stats_dict[key]
        total_length = sum(len(cluster) for cluster in clusters)
        cluster_length = -1
        num_clusters = len(clusters)
        for cluster in clusters:
            if stat in cluster:
                cluster_length = len(cluster)
                break
        return cluster_length, total_length, num_clusters

    def cluster_of(self, stat):
        key = (stat.dim1, stat.dim2, stat.ang_type)
        clusters = self._stats_dict[key]
        for i, cluster in enumerate(clusters):
            if stat in cluster:
                return i
        return -1

    def get_angle_stat_dims(self, dim0, dim1, ang_type):
        """
        Returns a list of pairs,`(dist, key)` ordered by increasing distance of the key to the query key.

        :param dim1: dim1 of the query key
        :param dim2: dim2 of the query key
        :param ang_type: angle_type of the query key.
        """
        available_stats = []
        for (k1, k2, k3) in self.keys():
            if k3 == ang_type:
                dist = m.sqrt((k1 - dim0) ** 2 + (k2 - dim1) ** 2)
                available_stats += [(dist, (k1, k2, k3))]

        available_stats.sort()
        return available_stats


class ConformationStats(object):
    def __init__(self, stats_file, clustered_angle_stats_file=None):
        if clustered_angle_stats_file is None:
            self.angle_stats = get_angle_stats(stats_file, refresh=True)
        else:
            self.angle_stats = ClusteredAngleStats(clustered_angle_stats_file)
        self.stem_stats = get_stem_stats(stats_file, refresh=True)
        self.fiveprime_stats = get_fiveprime_stats(stats_file, refresh=True)
        self.threeprime_stats = get_threeprime_stats(stats_file, refresh=True)
        self.loop_stats = get_loop_stats(stats_file, refresh=True)

        self.constrained_stats = c.defaultdict(list)

    def constrain_stats(self, constraint_file):
        '''
        Constrain the statistics for certain regions of the molecule. This
        is created for use with the JAR3D annotations for loop regions.

        :param constraint_file: A file containing the allowed statistics for
                                a particular loop.
        :return: Nothing
        '''
        warnings.warn("Function does nothing!!!")
        raise NotImplementedError("TODO")

    def sample_stats(self, bg, elem, min_entries=10):
        '''
        Return a set of statistics compatible with this element.

        :param bg: The graph representation we're using.
        :param elem: The name of the element
        :return: A list of compatible statistics
        '''
        if elem in self.constrained_stats:
            return self.constrained_stats[elem]

        dims = bg.get_node_dimensions(elem)

        if elem[0] == 's':
            dims = [dims]
            stats = self.stem_stats
            if stats[dims[0]]:
                return stats[dims[0]]
            else:
                raise LookupError("No stats for element {} with dimensions {}. Stats keys are {}".format(
                    elem, dims[0], sorted(stats.keys())))
        elif elem[0] == 'i' or elem[0] == 'm':
            stats = self.angle_stats

            ang_type = bg.get_angle_type(elem)
            try:
                if isinstance(stats, ClusteredAngleStats):
                    dims = stats.get_angle_stat_dims(
                        dims[0], dims[1], ang_type)
                else:  # assert isinstance(stats, defaultdict)
                    dims = get_angle_stat_dims(dims[0], dims[1],
                                               ang_type, min_entries=min_entries)
            except IndexError:
                print("Error in sample_stats:", file=sys.stderr)
                print("elem:", elem, "dims:", dims,
                      "ang_type:", ang_type, file=sys.stderr)
                raise
        elif elem[0] == 'h':
            dims = dims[0]
            stats = self.loop_stats
            dims = get_one_d_stat_dims(dims, stats)
        elif elem[0] == 't':
            dims = dims[0]
            stats = self.threeprime_stats
            dims = get_one_d_stat_dims(dims, stats)

        elif elem[0] == 'f':
            dims = dims[0]
            stats = self.fiveprime_stats
            dims = get_one_d_stat_dims(dims, stats)

        all_stats = []
        for dim in dims:
            if len(all_stats) > min_entries:
                continue

            all_stats += stats[dim[-1]]

        if len(all_stats) == 0:
            msg = "No statistics for bulge {} with dims {}".format(elem, dims)
            if elem[0] in "mi":
                msg += " and ang_type {}".format(ang_type)
            raise LookupError(msg)

        return all_stats


class FilteredConformationStats(ConformationStats):
    def __init__(self, stats_file, filter_filename=None, filter_prob=1):
        """
        :param filter_prob: Return the filtered stats with this probability, else all stats.
                            Default=1 (100%)
        """
        super(FilteredConformationStats, self).__init__(stats_file)

        self.filtered = None
        self.filtered_stats = None
        self.filter_prob = filter_prob
        if filter_filename is not None:
            self.from_file(filter_filename)

    def from_file(self, filename):
        '''
        Read the statistics in from a file, with the following formatting::

            sampled i3 1X8W_A 4 27 31 101 104 ""
            sampled i2 3U5F_6 4 1348 1348 1365 1367 "cWW AG or UU"
            sampled i2 2QBG_B 4 1013 1014 1103 1103 "cWW AG or UU"

        '''
        self.filtered = dict()
        self.filtered_stats = c.defaultdict(list)

        with open(filename, 'r') as f:
            reader = csv.reader(f, delimiter=' ', quotechar='"')
            for row in reader:
                if not row:
                    continue  # Empty line
                elem_name = row[0]
                define_len = int(row[2])
                pdb_id = row[1]
                dims = tuple(map(int, row[3:5]))
                define = map(int, row[5:5 + define_len])

                # get filtered stats for each type of angle
                ang_types = [1, -1]
                for at in ang_types:
                    # for interior loops, the mirror loop should induce the same
                    # bend as if the direction of the stem was reversed
                    # TODO: make sure this is true
                    for stat in it.chain(self.angle_stats[(dims[0], dims[1], at)],
                                         self.angle_stats[(dims[1], dims[0], -at)]):
                        if stat.pdb_name == pdb_id:
                            # print stat.define, define
                            pass

                        if stat.pdb_name == pdb_id and stat.define == define:
                            #print >>sys.stderr, "found filtered stats:", stat.pdb_name, define
                            self.filtered_stats[(elem_name, at)] += [stat]

    def sample_stats(self, bg, elem):
        r = rand.random()
        if self.filtered_stats is not None and r <= self.filter_prob:
            ang_type = bg.get_angle_type(elem)
            if (elem, ang_type) in self.filtered_stats:
                if len(self.filtered_stats[(elem, ang_type)]) > 0:
                    return self.filtered_stats[(elem, ang_type)]
        # No filtered stats found. Return normal stats for this element.
        return super(FilteredConformationStats, self).sample_stats(bg, elem)


def get_conformation_stats(stats_file, angle_stats_file=None):
    if ConstructionStats.conf_stats is not None:
        return ConstructionStats.conf_stats
    else:
        return ConformationStats(stats_file, angle_stats_file)


def set_conformation_stats(conf_stats):
    ConstructionStats.conf_stats = conf_stats



from setuptools import setup, Extension
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError
import subprocess
import os
import itertools
import warnings
import logging

logging.basicConfig()
log = logging.getLogger(__file__)

ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError, IOError)



class CustomBuildExtCommand(build_ext):
    """
    build_ext command for use when numpy headers are needed.
    See: https://stackoverflow.com/a/42163080/5069869
    """
    def run(self):

        # Import numpy here, only when headers are needed
        import numpy

        # Add numpy headers to include_dirs
        self.include_dirs.append(numpy.get_include())

        # Call original build_ext command
        build_ext.run(self)

try: #If we are in a git-repo, get git-describe version.
    path = os.path.abspath(os.path.dirname(__file__))
    forgi_version = subprocess.check_output(["git", "describe", "--always"], universal_newlines=True).strip()
    try:
        subprocess.check_call(['git', 'diff-index', '--quiet', 'HEAD', '--'], universal_newlines=True)
    except subprocess.CalledProcessError:
        forgi_version+="+uncommited_changes"
    #Use a subclass of build_py from distutils to costumize the build.
    class build_py(_build_py):
        def run(self):
            """
            During building, adds a variable with the complete version (from git describe)
            to forgi/__init__.py.
            """
            outfile = self.get_module_outfile(self.build_lib, ["forgi"], "__init__")
            try:
                os.remove(outfile) #If we have an old version, delete it, so _build_py will copy the original version into the build directory.
            except:
                pass
            # Superclass build
            _build_py.run(self)
            # Apped the version number to init.py
            with open(outfile, "a") as of:
                of.write('\n__complete_version__ = "{}"'.format(forgi_version))
except: #Outside of a git repo, do nothing.
    build_py = _build_py

class BuildFailed(Exception):
    pass



extras = {"forgi.visual":["matplotlib>=2.0"],
          "development":["cython"],
          "classification":["scikit-learn"],
          "pdbechem":['beautifulsoup4>=4.6'],
          "tests":["ddt", "hypothesis", "nose2"]
         }
extras["all"]=list(itertools.chain(extras.values()))
setup_args = {
      "zip_safe":False,
      "cmdclass":{'build_py': build_py, 'build_ext': CustomBuildExtCommand},
      "name":'forgi',
      "version":'2.2.3',
      "description":'RNA Graph Library',
      "author":'Bernhard Thiel, Peter Kerpedjiev',
      "author_email":'thiel@tbi.univie.ac.at',
      "license":'GNU GPL 3.0',
      "url":'http://www.tbi.univie.ac.at/~pkerp/forgi/',
      "ext_modules": [Extension("forgi.threedee.utilities.cytvec",
                                sources=["forgi/threedee/utilities/cytvec.pyx",
                                         "forgi/threedee/utilities/broken_ml_core.cpp",
                                        ],
                                extra_compile_args=['-O3', "-std=c++11"],
                                language='c++' )],
      "packages":['forgi', 'forgi.graph', 'forgi.threedee',
                'forgi.threedee.model', 'forgi.utilities',
                'forgi.threedee.utilities',
                'forgi.threedee.classification',
                'forgi.threedee.classification._training',
                'forgi._k2n_standalone', 'forgi.threedee.visual',
                'forgi.visual', 'forgi.projection'],
      "package_data":{'forgi.threedee': ['data/*.pdb', 'data/stats/temp.stats', 'data/average_atom_positions.json', 'data/aminor_geometries.csv', 'data/aminor_params.json']},
      "data_files":[("", ["CREDITS", "LICENSE"])],
      "scripts":['examples/rnaConvert.py',
               'examples/describe_cg.py',
               'examples/compare_RNA.py',
               'examples/visualize_rna.py',
               'examples/pseudoknot_analyzer.py',
               'examples/forgi_config.py'],
      "install_requires":[
		'numpy>=1.10.0',
                'scipy>=0.19.1',
                'pandas>=0.20',
                'future',
                'networkx>=2.0',
                'biopython',
                'appdirs>=1.4',
                'logging_exceptions>=0.1.9',
                'cython'
	],
      "setup_requires":[
        'numpy>=1.10.0',
        'cython',
        'setuptools>=18'
      ],
      "extras_require":extras,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    "classifiers":[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',
        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
         ],
     }

setup(**setup_args)



#!/usr/bin/env python
from __future__ import print_function

import logging
import os.path
import sys
if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO
import string

import numpy as np

import forgi.graph.bulge_graph as fgb
import forgi.utilities.commandline_utils as fuc
import forgi.threedee.utilities.pdb as ftup

log = logging.getLogger(__name__)


def to_bg_or_cg_string(cg):
    if not hasattr(cg, "coords") or not cg.coords.is_filled:
        return cg.to_bg_string()
    else:
        return cg.to_cg_string()


def cg_or_bg_extention(cg):
    if not hasattr(cg, "coords") or not cg.coords.is_filled:
        return ".bg"
    else:
        return ".cg"


def bg_to_elem_string(cg):
    return fgb.BulgeGraph.to_dotbracket_string(cg) + "\n" + fgb.BulgeGraph.to_element_string(cg, with_numbers=True)


def to_pdb(cg):
    chains = ftup.rename_chains_for_pdb(cg.chains)
    f = StringIO()
    ftup.output_multiple_chains(chains.values(), f, "pdb")
    return f.getvalue()


class OutFiletype:
    def __init__(self, write_fun, extention_fun, rna_type):
        self.convert = write_fun
        self.get_extention = extention_fun
        self.rna_type = rna_type


FILETYPES = {
    "forgi": OutFiletype(to_bg_or_cg_string, cg_or_bg_extention, "any"),
    "bpseq": OutFiletype(fgb.BulgeGraph.to_bpseq_string, lambda x: ".bpseq", "cg"),
    "fasta": OutFiletype(fgb.BulgeGraph.to_fasta_string, lambda x: ".fa", "cg"),
    "dotbracket": OutFiletype(fgb.BulgeGraph.to_dotbracket_string, lambda x: ".dotbracket", "any"),
    "neato": OutFiletype(fgb.BulgeGraph.to_neato_string, lambda x: ".neato", "any"),
    "element_string": OutFiletype(bg_to_elem_string, lambda x: ".element_string", "any"),
    "pdb": OutFiletype(to_pdb, lambda x: ".pdb", "3d"),

}

parser = fuc.get_rna_input_parser(
    "Convert RNA files between different file formats.", "+")
parser.add_argument("-T", "--target-type", default="forgi", choices=FILETYPES.keys(),
                    type=str, help="The target file-type to convert into.")
parser.add_argument("--to-file", action="store_true",
                    help="Store the converted RNA in files instead of printing them to stdout. "
                         "The file-name will be the RNA's name (if present), otherwise 'rna001' etc.")
parser.add_argument("--filename", type=str,
                    help="If this is present, --to-file will automatically be true."
                         "A target filename (or path) without extention. "
                         "If it is a filename, use the given filename instead of the RNA's name. "
                         "If more than one input-RNA is present, appends automatically a increasing number."
                         "If it is a directory, create files in this directory.")
parser.add_argument("-f", "--force", action="store_true",
                    help="Overwrite files, if they already exist. Note: In case of race conditions, "
                         "files could be overwritten even if this flag is not provided.")
parser.add_argument("--refold-missing", action="store_true",
                    help="Only used for conversion from PDB/ MMCIF to secondary structure formats."
                         "Requires the ViennaRNA library with python bindings installed."
                         "With this option, missing residues (i.e.residues without 3D coordinates)"
                         "are included in the output and their secondary structure is predicted."
                         "This prediction is done with RNAfold on the whole sequence using the "
                         "secondary structure from the PDB as hard constraints.")


if __name__ == "__main__":
    args = parser.parse_args()
    with fuc.hide_traceback():
        cgs = fuc.cgs_from_args(
            args, rna_type=FILETYPES[args.target_type].rna_type)

    if args.filename:
        args.to_file = True
        if os.path.isdir(args.filename):
            directory = args.filename
            filename = None
            args.filename = None
        else:
            directory, filename = os.path.split(args.filename)
    else:
        filename = None
        directory = ""
    if args.refold_missing and args.target_type and FILETYPES[args.target_type].rna_type == "3d":
        raise ValueError("Option refold-missing cannot"
                         "be used for output files with 3D information.")

    for i, cg in enumerate(cgs):
        if not args.to_file and i > 0:
            print("\n\n========================================\n\n")
        if args.refold_missing:
            cg = fuc.with_missing_refolded(cg)
        target_str = FILETYPES[args.target_type].convert(cg)
        if args.to_file:
            if filename:
                if len(cgs) > 0:
                    fn = filename + "{:03d}".format(i + 1)
                else:
                    fn = filename
            else:
                if cg.name and cg.name != "untitled":
                    fn = cg.name
                else:
                    fn = "rna{:03d}".format(i + 1)
            fn += FILETYPES[args.target_type].get_extention(cg)
            fn = os.path.join(directory, fn)
        else:
            fn = "-"
        with fuc.open_for_out(fn, args.force) as outfile:
            print(target_str, file=outfile)
            if fn != "-":
                log.info("file {} written".format(fn))



#!/usr/bin/python
from __future__ import print_function, unicode_literals
from builtins import str
from builtins import range
import sys
import copy
import re
import logging
from collections import defaultdict

from logging_exceptions import log_to_exception, log_exception

import forgi.utilities.debug as fud
import forgi.graph.residue as fgr

log = logging.getLogger(__name__)


def parse_resid(mcann_resid):
    '''
    Read in an MC-Annotate formatted resid and return a tuple
    that can be used as an index into a Bio.PDB.Chain.
    '''
    parts = mcann_resid.split('.')
    if len(parts) == 1:
        return (' ', int(parts[0]), ' ')
    else:
        return (' ', int(parts[0]), parts[1])


def format_resid(pdb_resid):
    '''
    Convert a PDB.Chain.Residue id to an MC-Annotate formatted
    residue identifier.
    '''

    if pdb_resid[2] == ' ':
        return str(pdb_resid[1])
    else:
        return str(pdb_resid[1]) + "." + pdb_resid[2]


def parse_chain_base(chain_base):
    """
    Parse the string identifying a chain and a base in an MC-Annotate generated
    annotation file.

    As an example, the string 'A33' means residue 33 on chain 'A'.

    @param chain_base: The MC-Annotate identification string (i.e. 'A33')
    @return: A pair containing the chain id and residue number (i.e. ('A', 33))
    """
    log.debug("Chain_base is '{}'".format(chain_base))
    if (ord(chain_base[0]) >= ord('A') and ord(chain_base[0]) <= ord('Z')) or (ord(chain_base[0]) >= ord('a') and ord(chain_base[0]) <= ord('z')):
        # normal string
        chain = chain_base[0]
        base = chain_base[1:]
    else:
        # quoted string (i.e. ''33'')
        if chain_base[0] == '\'':
            end_quote_idx = chain_base.find('\'', 1)
            chain = chain_base[1:end_quote_idx]
            base = chain_base[end_quote_idx + 1:]
        else:
            # no chain identifier
            chain = ''
            base = chain_base
    log.debug("base is %s", base)
    return (chain, base)


def parse_base_pair_id(base_pair_id):
    """
    Separate the two chain/base identifiers present in the interaction section of
    an MC-Annotate output file.

    @param base_pair_id: The identifier string for the interacting nucleotides (i.e. 'A33-B45')
    @return: 4-tuple containing of the form (chain1, res1, chain2, res2) i.e. ('A', 33, 'B', '45')
    """
    # A number in single quotes or a letter, followed by a (potentially negative) number and
    # potentiallly by an insertion code.
    residue_pattern = r"(?:'\d'|[A-Za-z])-?\d+(?:\.[A-Za-z])?"

    parts = re.findall(residue_pattern, base_pair_id)
    if len(parts) != 2:
        e = ValueError(
            "Invalid interaction in the MC-Annotate file: %s" % base_pair_id)
        with log_to_exception(log, e):
            log.error("Regex matched the following parts: %s", parts)
        raise e
    if "-".join(parts) != base_pair_id:
        raise ValueError(
            "Invalid interaction in the MC-Annotate file: %s" % base_pair_id)

    log.debug("Parts are '{}'".format(parts))
    (from_chain, from_base) = parse_chain_base(parts[0].strip())
    (to_chain, to_base) = parse_chain_base(parts[1].strip())

    return (from_chain, from_base, to_chain, to_base)


def get_interacting_base_pairs(line):
    """
    Return the identification part of an interaction line of an MC-Annotate output file.

    @param line: The entire interaction line
    @return: The first part, containing the identifiers of the interacting entities (i.e. 'A33-B45')
    """
    line_parts = line.split(' ')
    return parse_base_pair_id(line_parts[0])


def iterate_over_residue_list(mcannotate_lines):
    """
    Generator function allowing the iteration over all of the lines in
    the 'Residue conformations' section of an MC-Annotate output file.

    @param mcannotate_lines: All of the lines of an MC-Annotate output file
    @return: Yield only lines in the 'Residue conformations' section
    """
    residue_conf_line = False
    for line in mcannotate_lines:
        if line.find('Residue conformations') == 0:
            residue_conf_line = True
            continue
        if line.find('Adjacent stacking') == 0:
            residue_conf_line = False
            continue
        if residue_conf_line:
            yield line


def iterate_over_interactions(mcannotate_lines):
    """
    Generator function for the iteration over lines in the 'Base-pairs' section of an
    MC-Annotate output file.

    @param mcannotate_lines: All of the lines of an MC-Annotate output file
    @return: Yield only lines in the 'Base-pairs' section
    """
    base_pair_line = False
    for line in mcannotate_lines:
        if line.find("Base-pairs ---") == 0:
            base_pair_line = True
            continue
        if line.find("Residue conformations") == 0:
            base_pair_line = False
            continue
        if base_pair_line:
            log.debug("A Basepair line is '{}'".format(line))
            try:
                (from_chain, from_base, to_chain,
                 to_base) = get_interacting_base_pairs(line)
            except ValueError as ve:
                log_exception(ve, logging.WARNING, with_stacktrace=False)
                continue

            yield line.strip()


def get_dotplot(lines):
    """docstring for get_dotplot"""
    residues = []
    residue_types = []
    bps = defaultdict(lambda: -1)
    bpseq_str = ""

    for line in iterate_over_residue_list(lines):
        parts = line.split(' ')
        residues.append(parse_chain_base(parts[0]))  # A tuple chain, id
        residue_types += [parts[2]]

    paired = set()
    for line in iterate_over_interactions(lines):
        parts = line.split(' ')
        #bond_type = parts[3]
        # if bond_type.find('Ww/Ww') >= 0 or bond_type.find('Ww/Ws') >= 0 or bond_type.find('Ws/Ww') >= 0:
        if ((line.find('Ww/Ww') >= 0 and (line.find('A-U') >= 0 or
                                          line.find('U-A') >= 0 or
                                          line.find('C-G') >= 0 or
                                          line.find('G-C') >= 0)) or
            (line.find('Ws/Ww') >= 0 and line.find('U-G') >= 0) or
                (line.find('Ww/Ws') >= 0 and line.find('G-U') >= 0)):
            # if bond_type.find('Ww/Ww') >= 0:
            # print line
            chain1, base1, chain2, base2 = parse_base_pair_id(parts[0])
            res1 = (chain1, base1)
            res2 = (chain2, base2)
            if res1 in paired or res2 in paired:
                if log.isEnabledFor(logging.WARNING):
                    if res1 in bps:
                        existing = "{} - {}".format(res1, residues[bps[res1]])
                    else:
                        existing = "{} - {}".format(res2, residues[bps[res2]])
                    log.warning(
                        "Base-triple encountered: Ignoring basepair %s - %s, because basepair %s exists", res1, res2, existing)
                continue

            paired.add(res1)
            paired.add(res2)
            try:
                bps[res1] = residues.index(res2)
                bps[res2] = residues.index(res1)
            except ValueError as e:
                with log_to_exception(log, e):
                    log.error(
                        "bps = %s, residues = %s, res1 = %s, res2 = %s", bps, residues, res1, res2)
                raise

    for i in range(len(residue_types)):
        bpseq_str += "%d %s %s\n" % (i + 1,
                                     residue_types[i], bps[residues[i]] + 1)
    seq_ids = _seqids_from_residue_map(residues)
    return bpseq_str, seq_ids


def _seqids_from_residue_map(residue_map):
    """
    Create the list of seq_ids from the list of MC-Annotate identifiers in the
    residue map.
    """
    seq_ids = []
    for i, r in enumerate(residue_map):
        log.debug("Residue %r", r)
        (from_chain, from_base) = r
        seq_ids += [fgr.RESID(from_chain, parse_resid(from_base))]
    log.debug("Seq_ids are %s", seq_ids)
    return seq_ids




cdef extern from "broken_ml_core.h":
    cdef cppclass Vect:
        Vect (double,double,double) except +
        Vect() except +
        double x,y,z

    Vect _get_broken_ml_dev_core(
                    Vect fixed_s_pos,
                    Vect fixed_s_vec,
                    Vect orig_coords0,
                    Vect orig_coords1,
                    Vect orig_twist,
                    Vect s_twist,
                    double r1, double u1, double v1,
                    double u, double v, double t)

    void testdot()



import math


def atan3(p1, p2):
    '''
    Return the angle between the two points in the
    range between 0 and 2 * pi.

    The angle is first obtained from math.atan2. The if it is
    less 0, it is added to 2 * pi.

    @param p1: The first point
    @param p2: The second point
    @return: The angle between the two points.
    '''
    a = math.atan2(p1, p2)
    if a < 0:
        return 2 * math.pi + a
    else:
        return a


def clock_angle(a1, a2):
    '''
    The amount one needs to rotate from angle a1
    to angle a2 in a clockwise direction. I.e.
    increasing a1 until it reaches a2.

    @param a1: The first angle.
    @param a2: The second angle.
    '''
    if a2 < a1:
        a2 += 2. * math.pi
    return a2 - a1



from builtins import str
import collections
import logging

log = logging.getLogger(__name__)
Resid_base = collections.namedtuple("complete_resid", ["chain", "resid"])


class RESID(Resid_base):
    # Thanks to https://stackoverflow.com/a/42146452 for hints on extending named_tuples
    def __repr__(self):
        return "Res(\"" + resid_to_str(self) + "\")"

    def __new__(cls, chain, resid=None):
        if resid is None:
            r = resid_from_str(chain)
            return r
        else:
            resid = (resid[0], int(resid[1]), str(resid[2]))
            if chain is not None:
                chain = str(chain)
            return Resid_base.__new__(cls, chain, resid)


Res = RESID  # So the repr can be used to create a RESID


def resid_to_str(resid):
    if resid.chain is not None:
        out = "{}:{}".format(resid.chain, resid.resid[1])
    else:
        out = str(resid.resid[1])
    if resid.resid[2] != " ":
        out += ".{}".format(resid.resid[2])
    return out


def resid_from_str(resstr):
    resstr = str(resstr)  # Make sure we use future's newstring on python2
    if ":" in resstr:
        chain, resid = resstr.split(":")
    else:
        resid = resstr
        log.debug("No chain given in string {!r}".format(resstr))
        chain = None
    if chain is not None:
        chain = str(chain)
    idparts = resid.split(".")
    if len(idparts) == 1:
        idparts.append(" ")
    return RESID(chain, (' ', int(idparts[0]), str(idparts[1])))


def resid_from_biopython(residue):
    if residue.parent is not None:
        return RESID(residue.parent.id, residue.id)
    else:
        return RESID(None, residue.id)
